from fhirclient.models.location import Location
from fhirclient.models.practitioner import Practitioner
from fhirclient.models.bundle import Bundle
from inspqcommun.fhir.clients.base_client import BaseClient
from requests import Response

import requests
import json
import logging

log = logging.getLogger(__name__)

class PractitionerClient(BaseClient):

    practitioner_endpoint = "{base_url}/Practitioner"
    practitioner_by_id_endpoint = practitioner_endpoint + "/{id}"

    def __init__(self, base_url=None, base_uri=None, token_header=None, validate_certs=True) -> None:
        super().__init__(base_url=base_url, base_uri=base_uri, token_header=token_header)
        self.validate_certs = validate_certs

    def search(self, identifier=None, role=None, family=None, given=None):
        params = {}
        if identifier is not None:
            params["identifier"] = identifier
        if role is not None:
            params["role"] = role
        if family is not None:
            params["family"] = family
        if given is not None:
            params["given"] = given

        response = requests.get(
            url=self.practitioner_endpoint.format(base_url=self.get_fhir_url()),
            headers=self.headers,
            params=params,
            verify=self.validate_certs)
        log.info("GET Practitioner?{} : {}".format("&".join([ key + "=" + params[key] for key in params ]), response.status_code))
        return response

    def get_by_id(self, practitioner_id=None):
        response = None
        if practitioner_id is not None:
            response = requests.get(
                url=self.practitioner_by_id_endpoint.format(base_url=self.get_fhir_url(), id=practitioner_id),
                headers=self.headers,
                verify=self.validate_certs
            )
        log.info("GET Organization/{} : {}".format(practitioner_id, response.status_code if response.status_code else 404))
        return response

    def extract_practitioner_bundle_from_response(self, practitioner_response: Response) -> Bundle:
        if practitioner_response.status_code == 200:
            content_dict = json.loads(practitioner_response.content)
            return Bundle(jsondict=content_dict)
        return None
    
    def extract_practitioner_from_response(self, response: Response) -> Practitioner:
        if response.status_code == 200:
            content_dict = json.loads(response.content)
            return Practitioner(jsondict=content_dict)


from fhirclient.models.operationoutcome import OperationOutcome
from fhirclient.models.coding import Coding
from inspqcommun.fhir.visitors.base import BaseVisitor

class OperationOutcomeVisitor(BaseVisitor):

    def __init__(self, fhir_resource=None) -> None:
        self.setFhirResource(fhir_resource if fhir_resource else OperationOutcome())

    def getFhirResource(self) -> OperationOutcome:
        return super().getFhirResource()

    def get_issue_count(self):
        return len(self.getFhirResource().issue)

    def get_issue_code(self, index=0):
        return self.getFhirResource().issue[index].code

    def get_issue_severity(self, index=0):
        return self.getFhirResource().issue[index].severity

    def get_issue_details(self, index=0) -> Coding:
        details = self._get_coding_par_system(self.getFhirResource().issue[index].details)
        return details if details else self.getFhirResource().issue[index].details[0]


from fhirclient.models.bundle import Bundle, BundleEntry
from inspqcommun.fhir.visitors.base import BaseVisitor

class BundleVisitor(BaseVisitor):

    def __init__(self, fhir_resource=None):
        self.entries = []
        self.setFhirResource(fhir_resource if fhir_resource else Bundle())

    def getFhirResource(self) -> Bundle:
        return super().getFhirResource()

    def add_entry(self, resource, response=None):
        entry = BundleEntry()
        entry.resource = resource
        if response:
            entry.response = response

        if not self.getFhirResource().entry:
            self.getFhirResource().entry = []
        self.getFhirResource().entry.append(entry)

    def count_entries(self, recurse=False, resource_type=None):
        return len(self.get_entries(recurse=recurse, resource_type=resource_type))

    def get_entries(self, recurse=False, resource_type=None):
        entries = []
        if self.getFhirResource().entry is None:
            return []
        if recurse:
            for entry in self.getFhirResource().entry:
                if resource_type is None or entry.resource.resource_name == resource_type:
                    entries.append(entry.resource)
                if type(entry.resource) is Bundle:
                    bundle_fhir = BundleVisitor(fhir_resource=entry.resource)
                    sub_entries = bundle_fhir.get_entries(recurse=recurse, resource_type=resource_type)
                    for e in sub_entries:
                        entries.append(e)
            return entries
        else:
            return self.getFhirResource().entry
        


from inspqcommun.fa.ressources.ressource import Ressource
from datetime import date
from typing import List

class Usager(Ressource):
    
    def __init__(self, **kwargs) -> None:
        self.identifiants : List[Usager.Identifiant] = self._get_list_of_values_from_kwargs("identifiants", Usager.Identifiant, **kwargs)
        self.prenom : str =  self._get_value_from_kwargs("prenom", defaultValue='', **kwargs)
        self.nom : str = self._get_value_from_kwargs("nom", defaultValue='', **kwargs)
        self.dateNaissance : date = self._get_value_from_kwargs("dateNaissance", **kwargs)
        self.sexe : str = self._get_value_from_kwargs("sexe", defaultValue='', **kwargs)
        self.adresse : Usager.Adresse = self._get_value_from_kwargs("adresse", Usager.Adresse, **kwargs)
        self.telephone : Usager.Telephone = self._get_value_from_kwargs("telephone", Usager.Telephone, **kwargs)
        self.contacts : List[Usager.Contact] = self._get_list_of_values_from_kwargs("contacts", Usager.Contact, **kwargs)
        
        super().__init__("Usager")            

    class Identifiant(Ressource):
        
        def __init__(self, **kwargs) -> None:
            self.type : str = self._get_value_from_kwargs("type", **kwargs)
            self.valeur : str = self._get_value_from_kwargs("valeur", **kwargs)
    
    class Adresse(Ressource):

        def __init__(self, **kwargs) -> None:
            self.adresse : str = self._get_value_from_kwargs("adresse", **kwargs)
            self.ville : str = self._get_value_from_kwargs("ville", **kwargs)
            self.province : str  = self._get_value_from_kwargs("province", **kwargs)
            self.pays : str = self._get_value_from_kwargs("pays", **kwargs)
            self.codePostal : str = self._get_value_from_kwargs("codePostal", **kwargs)
        
        def get_province_code(self) -> str:
            if self.province:
                if self.province.lower() == 'ontario':
                    return 'ON'
            return 'QC'

    class Contact(Ressource):

        def __init__(self, **kwargs) -> None:
            self.type : str = self._get_value_from_kwargs("type", **kwargs)
            self.prenom : str = self._get_value_from_kwargs("prenom", **kwargs)
            self.nom : str = self._get_value_from_kwargs("nom", **kwargs)

    class Telephone(Ressource):

        def __init__(self, **kwargs) -> None:
            self.numero : str = self._get_value_from_kwargs("numero", **kwargs)
            self.extension : str = self._get_value_from_kwargs("extension", **kwargs)


from inspqcommun.fa.configuration import Configuration
from inspqcommun.fhir.clients.base_client import BaseClient
from inspqcommun.fhir.clients.patient_client import PatientClient
from inspqcommun.fhir.clients.immunization_client import ImmunizationClient
from inspqcommun.fhir.clients.value_set_client import ValueSetClient
from inspqcommun.fhir.clients.medication_client import MedicationClient
from inspqcommun.fhir.clients.practitioner_client import PractitionerClient
from inspqcommun.fhir.clients.location_client import LocationClient
from inspqcommun.fhir.clients.organization_client import OrganizationClient
from inspqcommun.fa.validateurs.validateur_usager import ValidateurUsager, ValidateurUsagerPourAppariement
from inspqcommun.fa.validateurs.validateur_acte_vaccinal import ValidateurActeVaccinal
from inspqcommun.fa.convertisseurs.convertisseur_usager import ConvertisseurUsager
from inspqcommun.fa.convertisseurs.convertisseur_acte_vaccinal import ConvertisseurActeVaccinal
from inspqcommun.fa.chargeur_fichiers import ChargeurFichiers
from inspqcommun.fa.chargement_service import ChargementService

class Injecteur:

    def __init__(self, configuration: Configuration) -> None:
        self.configuration: Configuration = configuration
        self.base_client: BaseClient = None
        self.patient_client: PatientClient = None
        self.immunization_client: ImmunizationClient = None
        self.value_set_client: ValueSetClient = None
        self.medication_client: MedicationClient = None
        self.practitioner_client: PractitionerClient = None
        self.location_client: LocationClient = None
        self.organization_client: OrganizationClient = None
        self.validateur_usager: ValidateurUsager = None
        self.validateur_usager_pour_appariement: ValidateurUsagerPourAppariement = None
        self.validateur_acte_vaccinal: ValidateurActeVaccinal = None
        self.convertisseur_usager: ConvertisseurUsager = None
        self.convertisseur_acte_vaccinal: ConvertisseurActeVaccinal = None
        self.chargeur_fichiers: ChargeurFichiers = None
        self.chargement_service: ChargementService = None

    def get_base_client(self) -> BaseClient:
        if not self.base_client:
            self.base_client = BaseClient(base_url=self.configuration.get_fonctions_allegees_url(),
                                          base_uri=self.configuration.get_fonctions_allegees_uri(),
                                          token_header=self.__build_headers(),
                                          validate_certs=False)
        return self.base_client

    def get_patient_client(self) -> PatientClient:
        if not self.patient_client:
            self.patient_client = PatientClient(base_url=self.configuration.get_fonctions_allegees_url(), 
                                                base_uri=self.configuration.get_fonctions_allegees_uri(), 
                                                token_header=self.__build_headers(),
                                                validate_certs=False)
        return self.patient_client
    
    def get_immunization_client(self) -> ImmunizationClient:
        if not self.immunization_client:
            self.immunization_client = ImmunizationClient(base_url=self.configuration.get_fonctions_allegees_url(), 
                                                          base_uri=self.configuration.get_fonctions_allegees_uri(), 
                                                          token_header=self.__build_headers(),
                                                          validate_certs=False)
        return self.immunization_client
    
    def get_value_set_client(self) -> ValueSetClient:
        if not self.value_set_client:
            self.value_set_client = ValueSetClient(base_url=self.configuration.get_fonctions_allegees_url(), 
                                                   base_uri=self.configuration.get_fonctions_allegees_uri(), 
                                                   token_header=self.__build_headers(),
                                                   validate_certs=False)
        return self.value_set_client
    
    def get_medication_client(self) -> MedicationClient:
        if not self.medication_client:
            self.medication_client = MedicationClient(base_url=self.configuration.get_fonctions_allegees_url(), 
                                                      base_uri=self.configuration.get_fonctions_allegees_uri(), 
                                                      token_header=self.__build_headers(),
                                                      validate_certs=False)
        return self.medication_client
    
    def get_practitioner_client(self) -> PractitionerClient:
        if not self.practitioner_client:
            self.practitioner_client = PractitionerClient(base_url=self.configuration.get_fonctions_allegees_url(), 
                                                          base_uri=self.configuration.get_fonctions_allegees_uri(), 
                                                          token_header=self.__build_headers(),
                                                          validate_certs=False)
        return self.practitioner_client
    
    def get_location_client(self) -> LocationClient:
        if not self.location_client:
            self.location_client = LocationClient(base_url=self.configuration.get_fonctions_allegees_url(), 
                                                  base_uri=self.configuration.get_fonctions_allegees_uri(), 
                                                  token_header=self.__build_headers(),
                                                  validate_certs=False)
        return self.location_client

    def get_organization_client(self) -> OrganizationClient:
        if not self.organization_client:
            self.organization_client = OrganizationClient(base_url=self.configuration.get_fonctions_allegees_url(), 
                                                  base_uri=self.configuration.get_fonctions_allegees_uri(), 
                                                  token_header=self.__build_headers(),
                                                  validate_certs=False)
        return self.organization_client

    def get_validateur_usager(self) -> ValidateurUsager:
        if not self.validateur_usager:
            self.validateur_usager = ValidateurUsager(value_set_client=self.get_value_set_client())
        return self.validateur_usager
    
    def get_validateur_usager_pour_appariement(self) -> ValidateurUsagerPourAppariement:
        if not self.validateur_usager_pour_appariement:
            self.validateur_usager_pour_appariement = ValidateurUsagerPourAppariement()
        return self.validateur_usager_pour_appariement
    
    def get_validateur_acte_vaccinal(self) -> ValidateurActeVaccinal:
        if not self.validateur_acte_vaccinal:
            self.validateur_acte_vaccinal = ValidateurActeVaccinal(validateur_usager_pour_appariement=self.get_validateur_usager_pour_appariement(), 
                                                  value_set_client=self.get_value_set_client(),
                                                  medication_client=self.get_medication_client(),
                                                  practitioner_client=self.get_practitioner_client(),
                                                  location_client=self.get_location_client())
        return self.validateur_acte_vaccinal

    def get_convertisseur_usager(self) -> ConvertisseurUsager:
        if not self.convertisseur_usager:
            self.convertisseur_usager = ConvertisseurUsager()
        return self.convertisseur_usager
    
    def get_convertisseur_acte_vaccinal(self) -> ConvertisseurActeVaccinal:
        if not self.convertisseur_acte_vaccinal:
            self.convertisseur_acte_vaccinal = ConvertisseurActeVaccinal(convertisseur_usager=self.get_convertisseur_usager(), 
                                                                         patient_client=self.get_patient_client())
        return self.convertisseur_acte_vaccinal
    
    def get_chargeur_fichiers(self) -> ChargeurFichiers:
        if not self.chargeur_fichiers:
            self.chargeur_fichiers = ChargeurFichiers(self.configuration.get_files_path())
        return self.chargeur_fichiers
    
    def get_chargement_service(self) -> ChargementService:
        if not self.chargement_service:
            self.chargement_service = ChargementService(chargeur_fichiers=self.get_chargeur_fichiers(), 
                            validateur_usager=self.get_validateur_usager(), 
                            validateur_acte_vaccinal=self.get_validateur_acte_vaccinal(),
                            convertisseur_usager=self.get_convertisseur_usager(),
                            convertisseur_acte_vaccinal=self.get_convertisseur_acte_vaccinal(),
                            patient_client=self.get_patient_client(),
                            immunization_client=self.get_immunization_client())
        return self.chargement_service
            
    def __build_headers(self):
        return {"Authorization": self.configuration.get_authorization_header()}


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import requests
import json
from urllib.parse import urlencode

URL_TOKEN = "{url}/realms/{realm}/protocol/openid-connect/token"

def get_token(base_url, auth_realm, client_id,
              auth_username, auth_password, client_secret=None, validate_certs=True):
    auth_url = URL_TOKEN.format(url=base_url, realm=auth_realm)
    temp_payload = {
        'grant_type': 'password',
        'client_id': client_id,
        'client_secret': client_secret,
        'username': auth_username,
        'password': auth_password,
    }
    r = {}
    # Remove empty items, for instance missing client_secret
    payload = dict(
        (k, v) for k, v in temp_payload.items() if v is not None)
    try:
        resp = requests.post(url=auth_url, data=payload, verify=validate_certs)
        r = resp.json()
    except ValueError as e:
        raise KeycloakException(
            'API returned invalid JSON when trying to obtain access token from %s: %s'
            % (auth_url, str(e)))
    except Exception as e:
        raise KeycloakException('Could not obtain access token from %s: %s'
                            % (auth_url, str(e)))

    try:
        return r['access_token']
    except KeyError:
        raise KeycloakException(
            'Could not obtain access token from %s' % auth_url)

class KeycloakException(Exception):
    """ Keycloak Exception """



from fhirclient.models.patient import Patient, PatientContact
from fhirclient.models.codeableconcept import CodeableConcept
from fhirclient.models.extension import Extension
from fhirclient.models.identifier import Identifier
from fhirclient.models.humanname import HumanName
from fhirclient.models.address import Address
from fhirclient.models.contactpoint import ContactPoint

from inspqcommun.fhir.visitors.base import BaseVisitor

class PatientVisitor(BaseVisitor):
    
    NAM_IDENTIFIER_EXTENSION_URL = "http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/extensions/#patient/healthcardorigin"
    NAM_IDENTIFIER_SYSTEM = "http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/vocabulary/identifierType?code=NAM"
    NIU_IDENTIFIER_SYSTEM = "http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/vocabulary/identifierType?code=NIUU"
    AUCUN_IDENTIFIER_SYSTEM = "http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/vocabulary/identifierType?code=AUCUN"
    MATCHRAMQ_EXTENSION_URL = "http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/extensions/#patient/matchramq"

    def __init__(self, fhir_resource: Patient=None):
        self.setFhirResource(fhir_resource=fhir_resource if fhir_resource else Patient())

    def getFhirResource(self) -> Patient:
        return super().getFhirResource() 

    def set_id(self, id):
        self.getFhirResource().id = id

    def get_id(self, patient=None):
        if patient is None:
            patient = self.fhir_resource
        return patient.id

    def get_matchramq(self) -> bool:
        if self.getFhirResource() and self.getFhirResource().extension:
            matchRamq = self.__trouver_extension_par_url(url=self.MATCHRAMQ_EXTENSION_URL, extensions=self.getFhirResource().extension)
            if matchRamq:
                return matchRamq.valueBoolean
        return False

    def set_active(self, active: bool):
        self.getFhirResource().active = active

    def get_active(self):
        return self.getFhirResource().active

    def set_name(self, given_name=None, family_name=None):
        if not given_name and not family_name:
            self.getFhirResource().name = None
        else:
            if not self.getFhirResource().name:
                self.getFhirResource().name = []
                name = HumanName()
                name.given = [given_name] if given_name else None
                name.family = [family_name] if family_name else None
                self.getFhirResource().name.append(name)

    def get_given_name(self, index=0):
        return self.getFhirResource().name[index].given[0] if self.getFhirResource().name[index] and self.getFhirResource().name[index].given else None

    def get_family_name(self, index=0):
        return self.getFhirResource().name[index].family[0] if self.getFhirResource().name and self.getFhirResource().name[index].family else None

    def set_gender(self, gender: str):
        self.getFhirResource().gender = gender

    def get_gender(self):
        return self.getFhirResource().gender

    def set_nam(self, nam:str=None):
        self.__definir_identifiant_par_system(valeur=nam, code='NAM', system=self.NAM_IDENTIFIER_SYSTEM)

    def get_nam(self) -> Identifier:
        if self.getFhirResource() and self.getFhirResource().identifier:
            for identifier in self.getFhirResource().identifier:
                if identifier.system == self.NAM_IDENTIFIER_SYSTEM:
                    return identifier
        return None

    def set_niu(self, niu:str=None):
        self.__definir_identifiant_par_system(valeur=niu, code='NIUU', system=self.NIU_IDENTIFIER_SYSTEM)
        if not self.getFhirResource().extension:
            self.getFhirResource().extension = []
        matchRamq = self.__trouver_extension_par_url(url=self.MATCHRAMQ_EXTENSION_URL, extensions=self.getFhirResource().extension)
        if matchRamq:
            matchRamq.valueBoolean = niu is not None
        else:
            matchRamq = Extension()
            matchRamq.url = self.MATCHRAMQ_EXTENSION_URL
            matchRamq.valueBoolean = niu is not None
            self.getFhirResource().extension.append(matchRamq)

    def __trouver_extension_par_url(self, url: str, extensions) -> Extension:
        if extensions:
            for extension in extensions:
                if extension.url == url:
                    return extension
        return None

    def get_niu(self) -> Identifier:
        if self.getFhirResource() and self.getFhirResource().identifier:
            for identifier in self.getFhirResource().identifier:
                if identifier.system == self.NIU_IDENTIFIER_SYSTEM:
                    return identifier
        return None

    def set_aucun_identifiant(self):
        self.set_nam()
        self.set_niu()
        self.__definir_identifiant_par_system(code='AUCUN', system=self.AUCUN_IDENTIFIER_SYSTEM)

    def set_birth_date(self, birth_date):
        self.getFhirResource().birthDate = self.str_date_to_fhir_date(birth_date)

    def get_birth_date(self):
        return self.fhir_date_to_str_date(fhir_date=self.getFhirResource().birthDate)

    def set_address(self, address_line=None, address_city=None, address_state=None, address_postal_code=None, address_country=None):
        if not address_line and not address_city and not address_state and not address_country and not address_postal_code:
            self.getFhirResource().address = None
        else:
            address = Address()
            address.city = address_city
            address.country = address_country
            address.state = address_state
            address.postalCode = address_postal_code
            address.line = [address_line]
            self.getFhirResource().address = [address]

    def get_address(self) -> Address:
        if self.getFhirResource() and self.getFhirResource().address:
            return self.getFhirResource().address[0]
        else:
            return None

    def set_mother_name(self, mother_given_name=None, mother_family_name=None):
        self.__definir_contact_par_type(type='MERE', given_name=mother_given_name, family_name=mother_family_name)

    def get_mother_name(self) -> HumanName:
        if self.getFhirResource() and self.getFhirResource().contact:
            for contact in self.getFhirResource().contact:
                for relationship in contact.relationship:
                    for coding in relationship.coding:
                        if coding.code == 'MERE' and coding.system == PatientVisitor.DEFAULT_CODING_SYSTEM and coding.version == PatientVisitor.DEFAULT_CODING_VERSION:
                            return contact.name
        return None
        
    def set_father_name(self, father_given_name=None, father_family_name=None):
        self.__definir_contact_par_type(type='PERE', given_name=father_given_name, family_name=father_family_name)
        
    def get_father_name(self):
        if self.getFhirResource() and self.getFhirResource().contact:
            for contact in self.getFhirResource().contact:
                for relationship in contact.relationship:
                    for coding in relationship.coding:
                        if coding.code == 'PERE' and coding.system == PatientVisitor.DEFAULT_CODING_SYSTEM and coding.version == PatientVisitor.DEFAULT_CODING_VERSION:
                            return contact.name
        return None

    def set_phone_number(self, phone_number: str=None):
        if phone_number:
            self.getFhirResource().telecom = []
            telecom = ContactPoint()
            telecom.system = 'phone'
            telecom.value = phone_number
            self.getFhirResource().telecom.append(telecom)
        else:
            self.getFhirResource().telecom = None
         
    def get_phone_number(self) -> str:
        if self.getFhirResource() and self.getFhirResource().telecom:
            for telecom in self.getFhirResource().telecom:
                if telecom.system == 'phone':
                    return telecom.value
        return None

    def __definir_identifiant_par_system(self, code: str, system:str, valeur:str=None):

        if self.getFhirResource().identifier:
            identifier_existant = None
            for identifier in self.getFhirResource().identifier:
                if identifier.system == system:
                    identifier_existant=identifier
        
            if identifier_existant:
                self.getFhirResource().identifier.remove(identifier_existant)
        
        if valeur or system == self.AUCUN_IDENTIFIER_SYSTEM:
            if not self.getFhirResource().identifier:
                self.getFhirResource().identifier = []
            
            self.getFhirResource().identifier.append(self.__creer_identifier(code=code, valeur=valeur, system=system, coding_system=self.DEFAULT_CODING_SYSTEM, coding_version=self.DEFAULT_CODING_VERSION))
    
    def __creer_identifier(self, code: str, system: str, coding_system, coding_version, valeur:str=None):
        identifier = Identifier()

        if (system == self.NAM_IDENTIFIER_SYSTEM):
            healthcardorigin = Extension()
            healthcardorigin.url = self.NAM_IDENTIFIER_EXTENSION_URL
            healthcardorigin.valueCodeableConcept = self._to_codeable_concept(code='QC', coding_system=coding_system, coding_version=coding_version)
            identifier.extension = [healthcardorigin]

        identifier.type = self._to_codeable_concept(code=code, coding_system=coding_system, coding_version=coding_version)
        identifier.system = system
        identifier.value = valeur
        
        return identifier
    
    def __definir_contact_par_type(self, type:str, given_name:str=None, family_name:str=None):
        if not self.getFhirResource().contact:
            self.getFhirResource().contact = []

        contact: PatientContact
        relationship: CodeableConcept
        contact_existant: PatientContact = None

        for contact in self.getFhirResource().contact:
            for relationship in contact.relationship:
                if self._get_coding_par_system(relationship).code == type:
                    contact_existant = contact
        
        if contact_existant:
            self.getFhirResource().contact.remove(contact_existant)
        
        if given_name and family_name:
            self.getFhirResource().contact.append(self.__creer_contact(given_name=given_name, family_name=family_name, type=type))
    
    def __creer_contact(self, given_name:str, family_name:str, type:str) -> PatientContact:
        contact = PatientContact()
        contact.relationship = [ self._to_codeable_concept(code=type) ]
        contact.name = HumanName()
        contact.name.given = [given_name]
        contact.name.family = [family_name]
        return contact


from fhirclient.models.medication import Medication
from inspqcommun.fhir.clients.base_client import BaseClient
from requests import Response
import requests
import json
import logging

log = logging.getLogger(__name__)

class MedicationClient(BaseClient):

    medication_endpoint = "{base_url}/Medication"
    get_by_trade_name_endpoint = medication_endpoint + "/{trade_name}"

    def __init__(self, base_url=None, base_uri=None, token_header=None, validate_certs=True) -> None:
        super().__init__(base_url=base_url, base_uri=base_uri, token_header=token_header)
        self.validate_certs = validate_certs

    def get_by_trade_name(self, trade_name):
        response = requests.get(
            url=self.get_by_trade_name_endpoint.format(base_url=self.get_fhir_url(), trade_name=trade_name),
            headers=self.headers,
            verify=self.validate_certs)
        log.info("GET Medication/{} : {}".format(trade_name, response.status_code))
        return response
    
    def extract_medication_from_response(self, medication_response: Response) -> Medication:
        if medication_response.status_code == 200:
            medication_dict = json.loads(medication_response.content)
            return Medication(jsondict=medication_dict)
        else:
            return None
            


from fhirclient.models.extension import Extension
from fhirclient.models.fhirdate import FHIRDate
from fhirclient.models.codeableconcept import CodeableConcept
from fhirclient.models.coding import Coding
from datetime import datetime

class BaseVisitor():
    DEFAULT_CODING_SYSTEM = 'http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/vocabulary'
    DEFAULT_CODING_VERSION = "1.0.0"

    __fhir_resource = None

    def getFhirResource(self):
        return self.__fhir_resource
    
    def setFhirResource(self, fhir_resource):
        self.__fhir_resource = fhir_resource

    def str_date_to_fhir_date(self, str_date):
        fhir_date  = FHIRDate()
        fhir_date.date = datetime.strptime(str_date, '%Y-%m-%d').date()
        return fhir_date

    def fhir_date_to_str_date(self, fhir_date):
        str_date = None
        if fhir_date is not None and fhir_date.date is not None:
            str_date = fhir_date.isostring
        return str_date

    def add_or_update_extension_to_extensions(self, extension, extensions):
        if extensions is None:
            extensions = [extension]
        else:
            ext_found = False
            for ext in extensions:
                if type(ext) is Extension:
                    if ext.url == extension.url:
                        extensions.remove(ext)
                        extensions.append(extension)
                        ext_found = True
                        break
                elif type(ext) is list:
                    for sub_ext in ext:
                        ext = self.add_or_update_extension_to_extensions(extension=sub_ext, extensions=ext)
            if not ext_found:
                extensions.append(extension)
        return extensions
    
    def _to_codeable_concept(self, code: str, display:str=None, coding_system: str=DEFAULT_CODING_SYSTEM, coding_version: str=DEFAULT_CODING_VERSION) -> CodeableConcept:
        codeableConcept = CodeableConcept()
        coding = Coding()
        coding.code = code
        if display:
            coding.display = display
        coding.system = coding_system
        coding.version = coding_version
        codeableConcept.coding = [coding]
        return codeableConcept
    
    def _get_coding_par_system(self, codeableConcept: CodeableConcept, coding_system:str=DEFAULT_CODING_SYSTEM, coding_version:str=DEFAULT_CODING_VERSION) -> Coding:
        coding: Coding
        for coding in codeableConcept.coding:
            if coding.system == coding_system and coding.version == coding_version:
                return coding
        return None
    
    def _creer_extension(self, url: str, **kwargs) -> Extension:
        extension = Extension()
        extension.url = url
        if "valueString" in kwargs:
            extension.valueString = kwargs["valueString"]
        elif "valueDate" in kwargs:
            extension.valueDate = kwargs["valueDate"]
        elif "valueCodeableConcept" in kwargs:
            extension.valueCodeableConcept = kwargs["valueCodeableConcept"]
        elif "valueInteger" in kwargs:
            extension.valueInteger = kwargs["valueInteger"]
        return extension


from inspqcommun.fa.validateurs.validateur_usager import ValidateurUsager, ValidateurUsagerPourAppariement
from inspqcommun.fa.ressources.acte_vaccinal import ActeVaccinal
from inspqcommun.fa.validateurs.validateur import Validateur
from inspqcommun.fhir.clients.value_set_client import ValueSetClient
from inspqcommun.fhir.clients.medication_client import MedicationClient
from inspqcommun.fhir.clients.practitioner_client import PractitionerClient
from inspqcommun.fhir.clients.location_client import LocationClient
from inspqcommun.fhir.visitors.value_set import ValueSetVisitor
from inspqcommun.fhir.visitors.medication import MedicationVisitor
from typing import List
import logging

log = logging.getLogger(__name__)

class ValidateurActeVaccinal(Validateur):

    USAGER_PRESENT_MESSAGE = "L'usager pour lequel l'acte vaccinal est administré doit être présent"
    AGENT_PRESENT_MESSAGE = "L'agent immunisant de l'acte vaccinal doit être présent"
    ID_LOT_PRESENT_MESSAGE = "L'id du lot doit être présent lors qu'un lot est spécifié"
    NUMERO_LOT_PRESENT_MESSAGE = "Le numéro du lot doit être présent lors qu'un lot est spécifié"
    DATE_ADMINISTRATION_PRESENTE_MESSAGE = "La date d'administration de l'acte vaccinal doit être présente"
    QUANTITE_VALEUR_PRESENTE_MESSAGE = "La quantité administrée doit être présente"
    ID_VACCINATEUR_PRESENT_MESSAGE = "L'id du vaccinateur doit être présent lorsqu'un vaccinateur est défini"
    ID_LIEU_VACCINATION_PRESENT_MESSAGE = "L'id du lieu de vaccination doit être présent lorsqu'un lieu de vaccination est défini"
    AUTEUR_COMMENTAIRE_PRESENT_MESSAGE = "L'auteur du commentaire doit être présent pour chaque commentaire"
    DATE_COMMENTAIRE_PRESENTE_MESSAGE = "La date du comemntaire doit être présente pour chaque commentaire"
    TEXTE_COMMENTAIRE_PRESENT_MESSAGE = "Le texte du commentaire doit être présent pour chaque commentaire"

    NOM_COMMERCIAL_INEXISTANT_MESSAGE = "Le nom commercial {} n'existe pas"
    ID_LOT_INEXISTANT_POUR_NOM_COMMERCIAL_MESSAGE = "L'id du lot {} n'existe pas dans les lots du produit au nom commercial {}"
    NUMERO_LOT_INEXISTANT_POUR_NOM_COMMERCIAL_MESSAGE = "Le numéro du lot {} n'existe pas dans les lots du produit au nom commercial {}"
    AGENT_INEXISTANT_MESSAGE = "L'agent immunisant au code {} n'existe pas"
    UNITE_POSOLOGIQUE_INEXISTANTE_MESSAGE = "L'unité posologique au code {} n'existe pas"
    SITE_ADMINISTRATION_INEXISTANT_MESSAGE = "Le site d'administration au code {} n'existe pas"
    VOIE_ADMINISTRATION_INEXISTANTE_MESSAGE = "La voie d'administration au code {} n'existe pas"
    RAISON_ADMINISTRATION_INEXISTANTE_MESSAGE = "La raison d'administration au code {} n'existe pas"
    VACCINATEUR_INEXISTANT_MESSAGE = "Le vaccinateur qui a l'id {} n'existe pas"
    LIEU_VACCINATION_INEXISTANT_MESSAGE = "Le lieu de vaccination qui a l'id {} n'existe pas"

    def __init__(self, validateur_usager_pour_appariement: ValidateurUsagerPourAppariement, value_set_client: ValueSetClient, 
                 medication_client: MedicationClient, practitioner_client: PractitionerClient, location_client: LocationClient) -> None:
        super().__init__(value_set_client)
        self.__validateur_usager_pour_appariement = validateur_usager_pour_appariement
        self.__medication_client = medication_client
        self.__practitioner_client = practitioner_client
        self.__location_client = location_client

    def valider(self, acte_vaccinal: ActeVaccinal) -> List[str]:
        erreurs: List[str] = []

        erreurs += self.__validateur_usager_pour_appariement.valider(acte_vaccinal.usager)
        erreurs += self.__valider_champs_obligatoires(acte_vaccinal)
        erreurs += self.__valider_domaines_vocabulaire(acte_vaccinal)

        if erreurs:
            log.warn("L'acte vaccinal n'est pas valide", extra={"erreurs":";".join(erreurs)})
        else:
            log.info("L'acte vaccinal a été validé et n'a pas d'erreurs")

        return erreurs
    
    def __valider_champs_obligatoires(self, acte_vaccinal: ActeVaccinal) -> List[str]:
        erreurs: List[str] = []

        if not acte_vaccinal.usager:
            log.debug(self.USAGER_PRESENT_MESSAGE)
            erreurs.append(self.USAGER_PRESENT_MESSAGE)
        
        if not acte_vaccinal.agent:
            log.debug(self.AGENT_PRESENT_MESSAGE)
            erreurs.append(self.AGENT_PRESENT_MESSAGE)
        
        if acte_vaccinal.lot:
            if not acte_vaccinal.lot.id:
                log.debug(self.ID_LOT_PRESENT_MESSAGE)
                erreurs.append(self.ID_LOT_PRESENT_MESSAGE)
            if not acte_vaccinal.lot.numero:
                log.debug(self.NUMERO_LOT_PRESENT_MESSAGE)
                erreurs.append(self.NUMERO_LOT_PRESENT_MESSAGE)
            
        if not acte_vaccinal.dateAdministration:
            log.debug(self.DATE_ADMINISTRATION_PRESENTE_MESSAGE)
            erreurs.append(self.DATE_ADMINISTRATION_PRESENTE_MESSAGE)
        
        if acte_vaccinal.quantiteAdministree:
            if not acte_vaccinal.quantiteAdministree.quantite:
                log.debug(self.QUANTITE_VALEUR_PRESENTE_MESSAGE)
                erreurs.append(self.QUANTITE_VALEUR_PRESENTE_MESSAGE)
        
        if acte_vaccinal.vaccinateur:
            if not acte_vaccinal.vaccinateur.id:
                log.debug(self.ID_VACCINATEUR_PRESENT_MESSAGE)
                erreurs.append(self.ID_VACCINATEUR_PRESENT_MESSAGE)

        if acte_vaccinal.lieuVaccination:
            if not acte_vaccinal.lieuVaccination.id:
                log.debug(self.ID_LIEU_VACCINATION_PRESENT_MESSAGE)
                erreurs.append(self.ID_LIEU_VACCINATION_PRESENT_MESSAGE)

        if acte_vaccinal.commentaires:
            for commentaire in acte_vaccinal.commentaires:
                if not commentaire.auteur:
                    log.debug(self.AUTEUR_COMMENTAIRE_PRESENT_MESSAGE)
                    erreurs.append(self.AUTEUR_COMMENTAIRE_PRESENT_MESSAGE)
                if not commentaire.date:
                    log.debug(self.DATE_COMMENTAIRE_PRESENTE_MESSAGE)
                    erreurs.append(self.DATE_COMMENTAIRE_PRESENTE_MESSAGE)
                if not commentaire.texte:
                    log.debug(self.TEXTE_COMMENTAIRE_PRESENT_MESSAGE)
                    erreurs.append(self.TEXTE_COMMENTAIRE_PRESENT_MESSAGE)
        
        return erreurs
    
    def __valider_domaines_vocabulaire(self, acte_vaccinal: ActeVaccinal) -> List[str]:
        erreurs: List[str] = []

        if acte_vaccinal.usager:
            if acte_vaccinal.usager.identifiants:
                for identifiant in acte_vaccinal.usager.identifiants:
                    if not self._code_present_dans_domaine_vocabulaire(identifiant.type, self._value_set_client.get_identifiertype()):
                        log.debug(ValidateurUsager.TYPE_IDENTIFIANT_INEXISTANT_MESSAGE.format(identifiant.type, identifiant.valeur))
                        erreurs.append(ValidateurUsager.TYPE_IDENTIFIANT_INEXISTANT_MESSAGE.format(identifiant.type, identifiant.valeur))
            if acte_vaccinal.usager.contacts:
                for contact in acte_vaccinal.usager.contacts:
                    if not self._code_present_dans_domaine_vocabulaire(contact.type, self._value_set_client.get_contacttype()):
                        log.debug(ValidateurUsager.TYPE_CONTACT_INEXISTANT_MESSAGE.format(contact.type, contact.prenom, contact.nom))
                        erreurs.append(ValidateurUsager.TYPE_CONTACT_INEXISTANT_MESSAGE.format(contact.type, contact.prenom, contact.nom))
                    
        if acte_vaccinal.nomCommercial:
            response = self.__medication_client.get_by_trade_name(trade_name=acte_vaccinal.nomCommercial)
            if not response.status_code == 200:
                log.debug(self.NOM_COMMERCIAL_INEXISTANT_MESSAGE.format(acte_vaccinal.nomCommercial))
                erreurs.append(self.NOM_COMMERCIAL_INEXISTANT_MESSAGE.format(acte_vaccinal.nomCommercial))
            else:
                if acte_vaccinal.lot:
                    medicationVisitor = MedicationVisitor(self.__medication_client.extract_medication_from_response(medication_response=response))
                    if acte_vaccinal.lot.id and not medicationVisitor.exists_lot_id_in_lots(acte_vaccinal.lot.id):
                        log.debug(self.ID_LOT_INEXISTANT_POUR_NOM_COMMERCIAL_MESSAGE.format(acte_vaccinal.lot.id, acte_vaccinal.nomCommercial))
                        erreurs.append(self.ID_LOT_INEXISTANT_POUR_NOM_COMMERCIAL_MESSAGE.format(acte_vaccinal.lot.id, acte_vaccinal.nomCommercial))
                    if  acte_vaccinal.lot.numero and not medicationVisitor.exists_lot_number_in_lots(acte_vaccinal.lot.numero):
                        log.debug(self.NUMERO_LOT_INEXISTANT_POUR_NOM_COMMERCIAL_MESSAGE.format(acte_vaccinal.lot.numero, acte_vaccinal.nomCommercial))
                        erreurs.append(self.NUMERO_LOT_INEXISTANT_POUR_NOM_COMMERCIAL_MESSAGE.format(acte_vaccinal.lot.numero, acte_vaccinal.nomCommercial))
            
        if acte_vaccinal.agent and not self._code_present_dans_domaine_vocabulaire(acte_vaccinal.agent, self._value_set_client.get_agent()):
            log.debug(self.AGENT_INEXISTANT_MESSAGE.format(acte_vaccinal.agent))
            erreurs.append(self.AGENT_INEXISTANT_MESSAGE.format(acte_vaccinal.agent))
            
        if acte_vaccinal.quantiteAdministree:
            if acte_vaccinal.quantiteAdministree.unite and not self._code_present_dans_domaine_vocabulaire(acte_vaccinal.quantiteAdministree.unite, self._value_set_client.get_dosageunit()):
                log.debug(self.UNITE_POSOLOGIQUE_INEXISTANTE_MESSAGE.format(acte_vaccinal.quantiteAdministree.unite))
                erreurs.append(self.UNITE_POSOLOGIQUE_INEXISTANTE_MESSAGE.format(acte_vaccinal.quantiteAdministree.unite))
        
        if acte_vaccinal.voieAdministration and not self._code_present_dans_domaine_vocabulaire(acte_vaccinal.voieAdministration, self._value_set_client.get_administrationroute()):
            log.debug(self.VOIE_ADMINISTRATION_INEXISTANTE_MESSAGE.format(acte_vaccinal.voieAdministration))
            erreurs.append(self.VOIE_ADMINISTRATION_INEXISTANTE_MESSAGE.format(acte_vaccinal.voieAdministration))
        
        if acte_vaccinal.siteAdministration and not self._code_present_dans_domaine_vocabulaire(acte_vaccinal.siteAdministration, self._value_set_client.get_administrationsite()):
            log.debug(self.SITE_ADMINISTRATION_INEXISTANT_MESSAGE.format(acte_vaccinal.siteAdministration))
            erreurs.append(self.SITE_ADMINISTRATION_INEXISTANT_MESSAGE.format(acte_vaccinal.siteAdministration))
        
        if acte_vaccinal.raisonAdministration and not self._code_present_dans_domaine_vocabulaire(acte_vaccinal.raisonAdministration, self._value_set_client.get_administrationreason()):
            log.debug(self.RAISON_ADMINISTRATION_INEXISTANTE_MESSAGE.format(acte_vaccinal.raisonAdministration))
            erreurs.append(self.RAISON_ADMINISTRATION_INEXISTANTE_MESSAGE.format(acte_vaccinal.raisonAdministration))
        
        if acte_vaccinal.vaccinateur and acte_vaccinal.vaccinateur.id:
            response = self.__practitioner_client.get_by_id(practitioner_id=acte_vaccinal.vaccinateur.id)
            if response.status_code == 404:
                log.debug(self.VACCINATEUR_INEXISTANT_MESSAGE.format(acte_vaccinal.vaccinateur.id))
                erreurs.append(self.VACCINATEUR_INEXISTANT_MESSAGE.format(acte_vaccinal.vaccinateur.id))
        
        if acte_vaccinal.lieuVaccination and acte_vaccinal.lieuVaccination.id:
            response = self.__location_client.get_by_id(location_id=acte_vaccinal.lieuVaccination.id)
            if response.status_code == 404:
                log.debug(self.LIEU_VACCINATION_INEXISTANT_MESSAGE.format(acte_vaccinal.lieuVaccination.id))
                erreurs.append(self.LIEU_VACCINATION_INEXISTANT_MESSAGE.format(acte_vaccinal.lieuVaccination.id))
        
        return erreurs


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import os
import json
from confluent_kafka import Producer, KafkaException, KafkaError

def obtenirConfigurationsProducteurDepuisVariablesEnvironnement():
    bootstrap_servers = os.environ['BOOTSTRAP_SERVERS'] if 'BOOTSTRAP_SERVERS' in os.environ else 'localhost:9092'
    
    config_kafka = {
        'bootstrap.servers': bootstrap_servers
    }
    if 'KAFKA_SECURITY_PROTOCOL' in os.environ:
        config_kafka['security.protocol'] = os.environ['KAFKA_SECURITY_PROTOCOL']
    if 'KAFKA_SSL_CA_LOCATION' in os.environ:
        config_kafka['ssl.ca.location'] = os.environ['KAFKA_SSL_CA_LOCATION']
    if 'KAFKA_SSL_CERTIFICATE_LOCATION' in os.environ:
        config_kafka['ssl.certificate.location'] = os.environ['KAFKA_SSL_CERTIFICATE_LOCATION']
    if 'KAFKA_SSL_KEY_LOCATION' in os.environ:
        config_kafka['ssl.key.location'] = os.environ['KAFKA_SSL_KEY_LOCATION']
    if 'KAFKA_SSL_KEY_PASSWORD' in os.environ:
        config_kafka['ssl.key.password'] = os.environ['KAFKA_SSL_KEY_PASSWORD']    

    return config_kafka    

def creerProducteur(config):
    producteur = Producer(**config)
    return producteur
"""
Fonction: publierMessage
Description: Cette fonction permet produire un message sur un topic Kafka
Paramètres:
    producteur: Producteur Kafka à utiliser
    message: Dictionnaire du message à publier:
        key: Clé pour publier le message
        value: Message à publier
    topic:Topic Kafka sur lequel publier le message
Retour:
    True si le message a bien été publié
"""    
def publierMessage(producteur, message, topic, logger=None):
    def callback_livraison(err, msg):
        if err:
            log_message = "Livraison du message impossible: {}".format(err)
            if logger is not None:
                logentry = {}
                logentry['topic'] = msg.topic()
                logentry['partition'] = msg.partition()
                logentry['offset'] = msg.offset()
                logentry['message_code'] = "ERREUR_PRODUCTEUR"
                logger.error(log_message, extra=logentry)
            
            else:
                print(log_message)
        else:
            log_message = "Le message a été livré. Topic {}, Partition {}, Offset {}".format(
                msg.topic(),
                msg.partition(),
                msg.offset())
            if logger is not None:
                logentry = {}
                logentry['topic'] = msg.topic()
                logentry['partition'] = msg.partition()
                logentry['offset'] = msg.offset()
                logentry['message_code'] = "CALLBACK_MESSAGE_PUBLIE"
                logger.info(log_message, extra=logentry)
            else:
                print(log_message)
                
    if logger is None:
        logging.basicConfig(level=logging.INFO)
        logger = logging.getLogger()
    try:
        
        key = encode_to_bytes(message['key']) if 'key' in message else None
        value = encode_to_bytes(message['value']) if 'value' in message else None
        producteur.produce(topic, value=value, key=key, callback=callback_livraison)
        logentry = {}
        logentry['topic'] = topic
        logentry['key'] = key.decode() if key is not None else ""
        logentry['value'] = value.decode() if value is not None else ""
        logentry['message_code'] = "MESSAGE_PUBLIE"
        log_message = "Message {} publié sur le topic {}".format(logentry['key'], topic)
        logger.info(log_message, extra=logentry)
    except BufferError as e:
        logentry = {}
        logentry['topic'] = topic
        logentry['key'] = key.decode() if key is not None else ""
        logentry['value'] = value.decode() if value is not None else ""
        logentry['message_code'] = "PRODUCTEUR_QUEUE_MESSAGE_PLEINE"
        log_message = "La queue de message est pleine. Erreur {}".format(str(e))
        logger.error(log_message, extra=logentry)
        return False
    except KafkaError as e:
        logentry = {}
        logentry['topic'] = topic
        logentry['key'] = key.decode() if key is not None else ""
        logentry['value'] = value.decode() if value is not None else ""
        logentry['message_code'] = "ERREUR_PRODUCTEUR"
        logentry['error_code'] = e.error_code
        logentry['reason'] = e.reason
        log_message = "Erreur Kafka: {}: {}".format(e.error_code, e.reason)
        logger.error(log_message, extra=logentry)
        return False
    except KafkaException as e:
        logentry = {}
        logentry['topic'] = topic
        logentry['key'] = key.decode() if key is not None else ""
        logentry['value'] = value.decode() if value is not None else ""
        logentry['message_code'] = "ERREUR_PRODUCTEUR"
        logentry['error_code'] = e.error_code
        logentry['reason'] = e.reason
        log_message = "Erreur Kafka: {}: {}".format(e.args[0].error_code, e.args[0].reason)
        logger.error(log_message, extra=logentry)
        return False
        
    producteur.flush()
    return True

def encode_to_bytes(value=None):
    if type(value) is str:
        return value.encode()
    elif type(value) is dict:
        return json.dumps(value).encode()
    else:
        return value


from inspqcommun.fa.ressources.ressource import Ressource

class Vaccinateur(Ressource):

    def __init__(self, **kwargs) -> None:
        super().__init__("Vaccinateur")
        self.id : int = self._get_value_from_kwargs("id", **kwargs)
        self.prenom : str = self._get_value_from_kwargs("prenom", **kwargs)
        self.nom : str = self._get_value_from_kwargs("nom", **kwargs)
        self.profession : str = self._get_value_from_kwargs("profession", **kwargs)
        self.numeroPermis : str = self._get_value_from_kwargs("numeroPermis", **kwargs)


from fhirclient.models.valueset import ValueSet
from inspqcommun.fhir.visitors.parameters import ParametersVisitor
from inspqcommun.fhir.clients.base_client import BaseClient
from requests import Response
import requests, json
import logging

log = logging.getLogger(__name__)

class ValueSetClient(BaseClient):

    valueset_endpoint = "{base_url}/ValueSet"
    valueset_vocabulary_domain_endpoint = valueset_endpoint + "/{domain}"
    valueset_city_endpoint = valueset_endpoint + "/$search-city"
    valueset_province_endpoint = valueset_endpoint + "/$provinces-by-country"

    def __init__(self, base_url=None, base_uri=None, token_header=None, validate_certs=True) -> None:
        super().__init__(base_url=base_url, base_uri=base_uri, token_header=token_header)
        self.validate_certs = validate_certs

    def get_valueset_by_domain(self, domain):
        response = requests.get(
            url=self.valueset_vocabulary_domain_endpoint.format(base_url=self.get_fhir_url(), domain=domain),
            headers=self.headers,
            verify=self.validate_certs
        )
        log.info("GET ValueSet/{} : {}".format(domain, response.status_code))
        return response
    
    def get_gender(self):
        return self.get_valueset_by_domain(domain="gender")

    def get_administrationsite(self):
        return self.get_valueset_by_domain(domain="administrationsite")
    
    def get_identifiertype(self):
        return self.get_valueset_by_domain(domain="identifiertype")
    
    def get_country(self):
        return self.get_valueset_by_domain(domain='country')
    
    def get_contacttype(self):
        return self.get_valueset_by_domain(domain='contacttype')
    
    def get_agent(self):
        return self.get_valueset_by_domain(domain='agent')
    
    def get_dosageunit(self):
        return self.get_valueset_by_domain(domain='dosageunit')
    
    def get_administrationroute(self):
        return self.get_valueset_by_domain(domain='administrationroute')
    
    def get_administrationreason(self):
        return self.get_valueset_by_domain(domain='administrationreason')
    
    def get_city_by_name(self, name: str, province_code:str='QC'):
        visitor = ParametersVisitor()
        if name:
            visitor.add_parameter(name, "city-name")
        if province_code:
            visitor.add_parameter(province_code, "province-code")
        
        response = requests.post(
            url=self.valueset_city_endpoint.format(base_url=self.get_fhir_url()),
            headers=self.headers,
            data=json.dumps(visitor.getFhirResource().as_json()),
            verify=self.validate_certs)
        log.info("POST ValueSet/$search-city - city-name={}&province-code={}: {}".format(name, province_code, response.status_code))
        return response        
    
    def get_provinces(self, country_code:str='CA'):
        visitor = ParametersVisitor()
        if country_code:
            visitor.add_parameter(country_code, 'country-code')
        
        response = requests.post(
            url=self.valueset_province_endpoint.format(base_url=self.get_fhir_url()),
            headers=self.headers,
            data=json.dumps(visitor.getFhirResource().as_json()),
            verify=self.validate_certs)
        log.info("POST ValueSet/$provinces-by-country - country-code={}: {}".format(country_code, response.status_code))
        return response
    
    def extract_value_set_from_response(self, valueset_response: Response) -> ValueSet:
        if valueset_response.status_code == 200:
            content_dict = json.loads(valueset_response.content)
            return ValueSet(jsondict=content_dict)
        else:
            return None


import os
import ast
from inspqcommun.identity.keycloak_tools import KeycloakEnvironment

DEFAULT_LOG_LEVEL: str = "INFO"
DEFAULT_FILES_PATH: str = None
DEFAULT_FA_BASE_URL: str = 'http://localhost:8089'
DEFAULT_FA_BASE_URI: str = '/fa-services'
DEFAULT_KEYCLOAK_URL: str = 'http://keycloak.url/auth'
DEFAULT_KEYCLOAK_AUTH_REALM: str = 'msss'
DEFAULT_KEYCLOAK_AUTH_USER: str = 'PERMISSIONS'
DEFAULT_KEYCLOAK_AUTH_PASSWORD: str = ''
DEFAULT_KEYCLOAK_AUTH_CLIENT_ID: str = "admin-cli"
DEFAULT_KEYCLOAK_ENABLED: str = "False"
DEFAULT_VALIDATE_CERTS: str = "true"
class Configuration:
    client_roles = ["fa-utilisateur", "fa-saisie"]
    def __init__(self, client_roles=None) -> None:
        if client_roles is not None:
            self.client_roles = client_roles

    def __str_to_bool(self, value: str) -> bool:
        bool_value = False
        try:
            bool_value = ast.literal_eval(value)
        except (SyntaxError, ValueError):
            return False
        return bool(bool_value)
    
    def get_log_level(self, default_value=DEFAULT_LOG_LEVEL) -> str:
        return os.environ.get("LOG_LEVEL", default_value)

    def get_files_path(self, default_value=DEFAULT_FILES_PATH) -> str:
        return os.environ.get("FILES_PATH", default_value)

    def get_fonctions_allegees_url(self, default_value=DEFAULT_FA_BASE_URL) -> str:
        return os.environ.get("FA_BASE_URL", default_value)

    def get_fonctions_allegees_uri(self, default_value=DEFAULT_FA_BASE_URI) -> str:
        return os.environ.get("FA_BASE_URI", default_value)
    
    def get_keycloak_base_url(self, default_value=DEFAULT_KEYCLOAK_URL) -> str:
        url = os.environ.get("KEYCLOAK_BASE_URL", default_value) 
        return url if url.endswith("/auth") else url + "/auth"
        
    def get_keycloak_auth_realm(self, default_value=DEFAULT_KEYCLOAK_AUTH_REALM) -> str:
        return os.environ.get("KEYCLOAK_AUTH_REALM", default_value)
    
    def get_keycloak_auth_user(self, default_value=DEFAULT_KEYCLOAK_AUTH_USER) -> str:
        return os.environ.get("KEYCLOAK_AUTH_USER", default_value)

    def get_keycloak_auth_password(self, default_value=DEFAULT_KEYCLOAK_AUTH_PASSWORD) -> str:
        return os.environ.get("KEYCLOAK_AUTH_PASSWORD", default_value)

    def get_keycloak_auth_client_id(self, default_value=DEFAULT_KEYCLOAK_AUTH_CLIENT_ID) -> str:
        return os.environ.get("KEYCLOAK_AUTH_CLIENT_ID", default_value)

    def get_keycloak_enabled(self, default_value=DEFAULT_KEYCLOAK_ENABLED) -> bool:
        return self.__str_to_bool(os.environ.get("KEYCLOAK_ENABLED", default_value))

    def get_validate_certs(self, default_value=DEFAULT_VALIDATE_CERTS) -> bool:
        return self.__str_to_bool(os.environ.get("VALIDATE_CERTS", default_value))
    
    def get_authorization_header(self) -> str:
        kcenv = KeycloakEnvironment(
            defaultAuthRealm=self.get_keycloak_auth_realm(),
            defaultAuthUser=self.get_keycloak_auth_user(),
            defaultAuthPassword=self.get_keycloak_auth_password(),
            defaultBaseKeycloakUrl=self.get_keycloak_base_url(),
            defaultKeycloakEnabled=self.get_keycloak_enabled(),
            defaultAuthClientId=self.get_keycloak_auth_client_id(),
            defaultValidateCert=self.get_validate_certs())
        headers = kcenv.authenticate(client_roles=self.client_roles)
        return headers.get('Authorization')




from fhirclient.models.location import Location
from fhirclient.models.fhirdate import FHIRDate
from fhirclient.models.period import Period
from fhirclient.models.extension import Extension

from inspqcommun.fhir.visitors.base import BaseVisitor

class LocationVisitor(BaseVisitor):
    
    RRSS_SYSTEM = 'https://pro.consultation.rrss.rtss.qc.ca'
    RRSS_MOT_CLE_EXTENSION_URL = 'http://www.santepublique.rtss.qc.ca/sipmi/rrss/1.0.0/extensions/#motscles'
    PERIOD_URL = "http://www.santepublique.rtss.qc.ca/sipmi/rrss/1.0.0/extensions/period"

    def __init__(self, fhir_resource=None) -> None:
        self.setFhirResource(fhir_resource=fhir_resource if fhir_resource else Location())

    def getFhirResource(self) -> Location:
        return super().getFhirResource()        

    def have_mot_cle(self, mot_cle=None):
        if mot_cle is not None and self.getFhirResource().extension is not None:
            for extension in self.getFhirResource().extension:
                if extension.url == self.RRSS_MOT_CLE_EXTENSION_URL and extension.valueString == mot_cle:
                    return True
        return False

    def set_name(self, name=None):
        self.getFhirResource().name = name

    def get_name(self):
        return self.getFhirResource().name

    def get_id(self):
        return self.getFhirResource().id

    def get_id_rrss(self):
        if self.getFhirResource().identifier is not None:
            for identifier in self.getFhirResource().identifier:
                if identifier.system == self.RRSS_SYSTEM:
                    return identifier.value
        return None

    def get_address_city(self):
        if self.getFhirResource().address:
            return self.getFhirResource().address.city
        return None

    def get_phones(self):
        phones = []
        if self.getFhirResource().telecom is None:
            return phones
        for telecom in self.getFhirResource().telecom:
            if telecom.system == "phone":
                phone = {}
                phone["value"] = telecom.value
                phone["use"] = telecom.use
                phones.append(phone)
        return phones

    def get_effective_from(self):
        period_ext = self.__get_period_ext()
        if period_ext is not None and period_ext.valuePeriod is not None:
                return period_ext.valuePeriod.start
        return None

    def get_effective_to(self):
        period_ext = self.__get_period_ext()
        if period_ext is not None and period_ext.valuePeriod is not None:
                return period_ext.valuePeriod.end
        return None

    def set_effective_from(self, effective_from=None):
        fhir_effective_from = None
        if type(effective_from) is str:
            fhir_effective_from = FHIRDate(jsonval=effective_from)
        elif type(effective_from) is FHIRDate:
            fhir_effective_from = effective_from
        period_ext = self.__get_period_ext()
        if period_ext is None or period_ext.valuePeriod is None:
            period = Period()
        else:
            period = period_ext.valuePeriod
        period.start = fhir_effective_from
        self.__set_period_ext(period=period)

    def set_effective_to(self, effective_to=None):
        fhir_effective_to = None
        if type(effective_to) is str:
            fhir_effective_to = FHIRDate(jsonval=effective_to)
        elif type(effective_to) is FHIRDate:
            fhir_effective_to = effective_to
        period_ext = self.__get_period_ext()
        if period_ext is None or period_ext.valuePeriod is None:
            period = Period()
        else:
            period = period_ext.valuePeriod
        period.end = fhir_effective_to
        self.__set_period_ext(period=period)

    def is_active(self):
        return self.getFhirResource().status == "active"
    
    def __get_period_ext(self):
        if self.getFhirResource().extension is not None:
            for ext in self.getFhirResource().extension:
                if ext.url == self.PERIOD_URL:
                    return ext
        return None

    def __set_period_ext(self, period=None):
        if self.getFhirResource().extension is None:
            self.getFhirResource().extension = []
        period_ext = self.__get_period_ext()
        if period_ext is None:
            period_ext = Extension()
            period_ext.url = self.PERIOD_URL
            self.getFhirResource().extension.append(period_ext)
        period_ext.valuePeriod = period
        
    


import requests
import json
from requests import Response
from fhirclient.models.patient import Patient
from fhirclient.models.identifier import Identifier
from fhirclient.models.bundle import Bundle
from fhirclient.models.parameters import Parameters
from fhirclient.models.fhirdate import FHIRDate
from inspqcommun.fhir.clients.base_client import BaseClient
from inspqcommun.fhir.visitors.parameters import ParametersVisitor
from inspqcommun.fhir.visitors.patient import PatientVisitor

import logging

log = logging.getLogger(__name__)

class PatientClient(BaseClient):

    base_url = ""
    patient_endpoint = "{0}/Patient"
    patient_id_endpoint = "{0}/Patient/{1}"
    patient_match_endpoint = "{0}/Patient/$match"
    patient_definir_niu_endpoint = patient_id_endpoint + "/$definir-niu"

    def __init__(self, base_url=None, base_uri=None, token_header=None, validate_certs=True) -> None:
        super().__init__(base_url=base_url,token_header=token_header, base_uri=base_uri)
        self.validate_certs = validate_certs
        if base_url is not None:
            self.base_url = base_url

    def create(self, patient: Patient):
        response = requests.post(
            url=self.patient_endpoint.format(self.get_fhir_url()),
            data=json.dumps(patient.as_json()),
            headers=self.headers,
            verify=self.validate_certs)
        log.info("POST Patient : {}".format(response.status_code))
        return response

    def update(self, patient: Patient):
        response = requests.put(
            url=self.patient_id_endpoint.format(self.get_fhir_url(), patient.id),
            data=json.dumps(patient.as_json()),
            headers=self.headers,
            verify=self.validate_certs)
        log.info("PUT Patient : {}".format(response.status_code))
        return response

    def get_by_id(self, patient_id=None):
        response = requests.get(url=self.patient_id_endpoint.format(self.get_fhir_url(), patient_id), headers=self.headers, verify=self.validate_certs)
        log.info("GET Patient/{} : {}".format(patient_id, response.status_code if response.status_code else 404))
        return response

    def match(self, patient: Patient):
        parametre = ParametersVisitor()
        parametre.add_parameter(patient)
        res = parametre.getFhirResource().as_json()
        response = requests.post(url=self.patient_match_endpoint.format(self.get_fhir_url()), data=json.dumps(res), headers=self.headers, verify=self.validate_certs)
        log.info("POST Patient/$match : {}".format(response.status_code))
        return response

    def search(self, identifier=None, nam=None, given=None, family=None, gender=None, birthdate=None):
        params = {}
        if identifier is not None:
            params['identifier'] = identifier
        elif nam is not None:
            params['identifier'] = "code=NAM|{}".format(nam)
        if given is not None:
            params['given'] = given
        if family is not None:
            params['family'] = family
        if gender is not None:
            params['gender'] = gender
        if birthdate is not None:
            birthdate_param = birthdate.isostring if type(birthdate) is FHIRDate else birthdate
            params['birthdate'] = birthdate_param

        response = requests.get(url=self.patient_endpoint.format(self.get_fhir_url()), params=params, headers=self.headers, verify=self.validate_certs)
        log.info("GET Patient?{} : {}".format("&".join([ key + "=" + str(params[key]) for key in params ]), response.status_code))
        return response

    def extract_patient_from_response(self, response: Response) -> Patient:
        content_dict = json.loads(response.content)
        if (content_dict['resourceType'] == 'Parameters'):
            params = Parameters(jsondict=content_dict)
            if params is not None and params.parameter is not None:
                for param in params.parameter:
                    if type(param.resource) is Patient:
                        return param.resource
        else:
            return Patient(jsondict=content_dict)
        
    def extract_patient_bundle_from_response(self, response: Response) -> Bundle:
        json_resource = json.loads(response.content.decode())
        if (json_resource['resourceType'] == 'Bundle'):
            return Bundle(jsondict=json.loads(response.content))
        return None
    
    def get_immunization_profile(self, patient_id: int):
        self.patient_id_endpoint.format(self.get_fhir_url(), patient_id)
        url = "{0}/$immunization-profile".format(self.patient_id_endpoint.format(self.get_fhir_url(), patient_id))
        response = requests.get(url=url, headers=self.headers, verify=self.validate_certs)
        return response

    def __trouver_niuu_de_patient(self, patient: Patient) -> Identifier:
        if patient.identifier:
            identifier: Identifier
            for identifier in patient.identifier:
                if identifier.system == PatientVisitor.NIU_IDENTIFIER_SYSTEM:
                    return identifier
        return None


from inspqcommun.fhir.visitors.value_set import ValueSetVisitor
from inspqcommun.fhir.clients.value_set_client import ValueSetClient
from requests import Response
import logging

log = logging.getLogger(__name__)

class Validateur:

    def __init__(self, value_set_client: ValueSetClient) -> None:
        self._value_set_client = value_set_client

    def _code_present_dans_domaine_vocabulaire(self, code: str, response: Response) -> bool:
        if response.status_code == 200:
            value_set = self._value_set_client.extract_value_set_from_response(response)
            return ValueSetVisitor(value_set).find_code_system_concept_by_code(code)
        else:
            log.error("Erreur lors de l'appel aux Fonctions allégées: {} {}".format(response.status_code, response.content))
            return False
    
    def _description_presente_dans_domaine_vocabulaire(self, description: str, response: Response) -> bool:
        if response.status_code == 200:
            value_set = self._value_set_client.extract_value_set_from_response(response)
            return ValueSetVisitor(value_set).find_code_system_concept_by_display(description)
        else:
            log.error("Erreur lors de l'appel aux Fonctions allégées: {} {}".format(response.status_code, response.content))
            return False


from fhirclient.models.parameters import Parameters, ParametersParameter
from fhirclient.models.resource import Resource
from inspqcommun.fhir.visitors.base import BaseVisitor

class ParametersVisitor(BaseVisitor):

    def __init__(self, fhir_resource=None):
        self.setFhirResource(fhir_resource if fhir_resource is not None else Parameters())

    def getFhirResource(self) -> Parameters:
        return super().getFhirResource()

    def add_parameter(self, valeur : any, name='resource'):
        parameter = ParametersParameter()
        parameter.name=name
        if isinstance(valeur, Resource):
            parameter.resource = valeur
        elif isinstance(valeur, str):
            parameter.valueString = valeur

        if self.getFhirResource().parameter is None:
            self.getFhirResource().parameter = []
        
        self.getFhirResource().parameter.append(parameter)
    
    def find_by_name(self, name='resource') -> ParametersParameter:
        if self.getFhirResource().parameter:
            parameter: ParametersParameter
            for parameter in self.getFhirResource().parameter:
                if parameter.name == name:
                    return parameter
        return None


from inspqcommun.fa.validateurs.validateur_usager import ValidateurUsager
from inspqcommun.fa.validateurs.validateur_acte_vaccinal import ValidateurActeVaccinal
from inspqcommun.fa.convertisseurs.convertisseur_usager import ConvertisseurUsager
from inspqcommun.fa.convertisseurs.convertisseur_acte_vaccinal import ConvertisseurActeVaccinal
from inspqcommun.fa.ressources.usager import Usager
from inspqcommun.fa.ressources.acte_vaccinal import ActeVaccinal
from inspqcommun.fa.chargeur_fichiers import ChargeurFichiers
from inspqcommun.fhir.clients.patient_client import PatientClient
from inspqcommun.fhir.clients.immunization_client import ImmunizationClient
from fhirclient.models.operationoutcome import OperationOutcome
from typing import List
import logging, json

log = logging.getLogger(__name__)

class ChargementService:

    def __init__(self, chargeur_fichiers : ChargeurFichiers, validateur_usager: ValidateurUsager, validateur_acte_vaccinal: ValidateurActeVaccinal,
                 convertisseur_usager : ConvertisseurUsager, convertisseur_acte_vaccinal : ConvertisseurActeVaccinal, patient_client: PatientClient,
                 immunization_client: ImmunizationClient):
        self.__chargeur_fichiers : ChargeurFichiers = chargeur_fichiers
        self.__validateur_usager : ValidateurUsager = validateur_usager
        self.__validateur_acte_vaccinal : ValidateurActeVaccinal = validateur_acte_vaccinal
        self.__convertisseur_usager : ConvertisseurUsager = convertisseur_usager
        self.__convertisseur_acte_vaccinal : ConvertisseurActeVaccinal = convertisseur_acte_vaccinal
        self.__patient_client : PatientClient = patient_client
        self.__immunization_client : ImmunizationClient = immunization_client
        self.nb_ressources_fichiers = 0
        self.nb_ressources_converties = 0
        self.nb_fichiers = 0
        self.nb_fichiers_valides = 0

    def charger_donnees_par_fonctions_allegees(self):
        tous_les_fichiers_sont_valides = True
        toutes_les_ressources_sont_valides = True
        toutes_les_ressources_ont_ete_chargees_avec_succes = True
        for ressources_a_charger in self.__valider_et_obtenir_ressources_dans_les_fichiers():
            for ressource in ressources_a_charger:
                if isinstance(ressource, Usager):
                    if not self.__charger_usager(usager=ressource):
                        toutes_les_ressources_ont_ete_chargees_avec_succes = False
                elif isinstance(ressource, ActeVaccinal):
                    if not self.__charger_actes_vaccinaux(acte_vaccinal=ressource):
                        toutes_les_ressources_ont_ete_chargees_avec_succes = False
        log.info("Nb fichiers: {0}, nb fichiers valides: {1}".format(self.nb_fichiers, self.nb_fichiers_valides))
        log.info("Nb ressources dans fichiers: {0}, nb ressource valides: {1}".format(self.nb_ressources_fichiers, self.nb_ressources_converties))
        if not toutes_les_ressources_ont_ete_chargees_avec_succes:
            log.error("Certaines ressources n'ont pas été chargées")
        if self.nb_fichiers_valides < self.nb_fichiers:
            tous_les_fichiers_sont_valides = False
            log.error("Certains fichiers ne sont pas valides")
        if self.nb_ressources_converties < self.nb_ressources_fichiers:
            toutes_les_ressources_sont_valides = False
            log.error("Certaines ressources dans les fichiers ne sont pas valides")
        return tous_les_fichiers_sont_valides and toutes_les_ressources_sont_valides and toutes_les_ressources_ont_ete_chargees_avec_succes

    def valider_fichier_donnees(self) -> bool:
        tous_les_fichiers_sont_valides = True
        toutes_les_ressources_sont_valides = True
        self.__valider_et_obtenir_ressources_dans_les_fichiers()
        log.info("Nb fichiers: {0}, nb fichiers valides: {1}".format(self.nb_fichiers, self.nb_fichiers_valides))
        log.info("Nb ressources dans fichiers: {0}, nb ressource valides: {1}".format(self.nb_ressources_fichiers, self.nb_ressources_converties))
        if self.nb_fichiers_valides < self.nb_fichiers:
            tous_les_fichiers_sont_valides = False
            log.error("Certains fichiers ne sont pas valides")
        if self.nb_ressources_converties < self.nb_ressources_fichiers:
            tous_les_fichiers_sont_valides = False
            log.error("Certaines ressources dans les fichiers ne sont pas valides")
        return tous_les_fichiers_sont_valides and toutes_les_ressources_sont_valides

    def __valider_et_obtenir_ressources_dans_les_fichiers(self):
        fichier_en_erreur: bool = False
        listes_ressources_a_charger = []

        for nom_fichier in self.__chargeur_fichiers.obtenir_noms_fichiers():
            ressources_a_charger = []
            validations_echouees: bool = False

            contenu_fichier = self.__chargeur_fichiers.obtenir_contenu_fichier_par_nom(nom_fichier)
            self.nb_fichiers += 1
            if contenu_fichier is not None:
                self.nb_fichiers_valides += 1
                for ressource in contenu_fichier:
                    self.nb_ressources_fichiers += 1
                    if "ressource" in ressource:
                        ressource_convertie = None
                        if ressource["ressource"] == "Usager":
                            ressource_convertie = self.__obtenir_ressource_convertie_si_valide(ressource, self.__convertisseur_usager, self.__validateur_usager)
                        if ressource["ressource"] ==  "ActeVaccinal":
                            ressource_convertie = self.__obtenir_ressource_convertie_si_valide(ressource, self.__convertisseur_acte_vaccinal, self.__validateur_acte_vaccinal)
                            
                        if ressource_convertie:
                            self.nb_ressources_converties += 1
                            ressources_a_charger.append(ressource_convertie)
                        else:
                            log.error("La ressource {} n'a pas pu être convertie".format(str(ressource)))
                            validations_echouees = True

                if not validations_echouees:
                    listes_ressources_a_charger.append(ressources_a_charger)
                else:
                    fichier_en_erreur = True
            else:
                fichier_en_erreur = True

        return listes_ressources_a_charger if not fichier_en_erreur else []

    def __obtenir_ressource_convertie_si_valide(self, ressource: str, convertisseur, validateur) -> any:
        ressource_convertie = convertisseur.fromJson(**ressource)
        erreurs = validateur.valider(ressource_convertie)
        if not erreurs:
            return ressource_convertie
        else:
            log.error("La ressource de type {} n'est pas valide: {}".format(type(ressource_convertie), ";".join(erreurs)))
            return None
    
    def __charger_usager(self, usager: Usager):
        chargement_en_succes = True
        patient = self.__convertisseur_usager.toFhir(usager)
        match_response = self.__patient_client.match(patient)
        if match_response.status_code == 204:
            create_response = self.__patient_client.create(patient)
            if create_response.status_code == 200:
                log.info("L'usager a été créé avec succès")
            else:
                outcome = OperationOutcome(json.loads(create_response.content))
                log.error("La création de l'usager {0} a eu des problèmes: {1}".format(str(usager.as_json()), json.dumps(outcome.as_json())))
                chargement_en_succes = False
        elif match_response.status_code == 200:
            # TODO supporter la modification de l'usager
            log.warn("L'usager n'a pas pu être créé car il existe déjà")
        else:
            outcome = OperationOutcome(json.loads(match_response.content))
            log.error("L'appariement de l'usager {0} a eu des problèmes: {1}".format(str(usager.as_json()),json.dumps(outcome.as_json())))
            chargement_en_succes = False
        return chargement_en_succes
    
    # TODO supporter la modification de l'acte vaccinal s'il existe (faire un appariement)
    def __charger_actes_vaccinaux(self, acte_vaccinal: ActeVaccinal):
        chargement_en_succes = True
        immunization = self.__convertisseur_acte_vaccinal.toFhir(acte_vaccinal)
        response = self.__immunization_client.create(immunization)
        if response.status_code == 201:
            log.info("L'acte vaccinal a été créé avec succès")
        else:
            outcome = OperationOutcome(json.loads(response.content))
            log.error("La création de l'acte vaccinal {0} n'a pas fonctionné: {1}".format(acte_vaccinal.as_json(), json.dumps(outcome.as_json())))
            chargement_en_succes = False
        return chargement_en_succes
        



from inspqcommun.fa.ressources.usager import Usager
from inspqcommun.fhir.clients.value_set_client import ValueSetClient
from inspqcommun.fa.validateurs.validateur import Validateur
from typing import List
import logging

log = logging.getLogger(__name__)

class ValidateurUsager(Validateur):
    TYPE_IDENTIFIANT_POUR_NIU = "NIUU"
    PRENOM_USAGER_PRESENT_MESSAGE = "Le prénom de l'usager doit être présent"
    NOM_USAGER_PRESENT_MESSAGE = "Le nom de l'usager doit être présent"
    DATE_NAISSANCE_USAGER_PRESENTE_MESSAGE = "La date de naissance de l'usager doit être présente"
    SEXE_USAGER_PRESENT_MESSAGE = "Le sexe de l'usager doit être présent"
    IDENTIFIANTS_OU_CONTACTS_USAGER_PRESENT_MESSAGE = "L'usager doit avoir au moins un identifiant ou un contact"
    TYPE_ET_VALEUR_IDENTIFIANTS_USAGER_PRESENTS_MESSAGE = "Tous les identifiants doivent avoir un type et une valeur"
    TYPE_ET_PRENOM_ET_NOM_CONTACTS_USAGER_PRESENTS_MESSAGE = "Tous les contacts doivent avoir un type, un prénom et un nom"

    TYPE_IDENTIFIANT_INEXISTANT_MESSAGE = "Le type d'identifiant {} pour l'identifiant {} n'est pas un type d'identifiant qui existe"
    SEXE_INEXISTANT_MESSAGE = "Le sexe {} n'est pas un sexe qui existe"
    VILLE_INEXISTANTE_MESSAGE = "La ville {} n'est pas une ville qui existe"
    PROVINCE_INEXISTANTE_MESSAGE = "La province {} n'est pas une province qui existe"
    PAYS_INEXISTANT_MESSAGE = "Le pays {} n'est pas un pays qui existe"
    TYPE_CONTACT_INEXISTANT_MESSAGE = "Le type de contact {} pour le contact {} {} n'est pas un type de contact qui existe"

    def __init__(self, value_set_client: ValueSetClient) -> None:
        super().__init__(value_set_client)

    def valider(self, usager: Usager) -> List[str]:
        erreurs: List[str] = []
        
        erreurs += self.__valider_champs_obligatoires(usager)
        erreurs += self.__valider_domaines_vocabulaire(usager)
        
        return erreurs

    def __valider_champs_obligatoires(self, usager: Usager) -> List[str]:
        erreurs: List[str] = []
        if not usager.prenom:
            log.debug(self.PRENOM_USAGER_PRESENT_MESSAGE)
            erreurs.append(self.PRENOM_USAGER_PRESENT_MESSAGE)
        if not usager.nom:
            log.debug(self.NOM_USAGER_PRESENT_MESSAGE)
            erreurs.append(self.NOM_USAGER_PRESENT_MESSAGE)
        if not usager.dateNaissance:
            log.debug(self.DATE_NAISSANCE_USAGER_PRESENTE_MESSAGE)
            erreurs.append(self.DATE_NAISSANCE_USAGER_PRESENTE_MESSAGE)
        if not usager.sexe:
            log.debug(self.SEXE_USAGER_PRESENT_MESSAGE)
            erreurs.append(self.SEXE_USAGER_PRESENT_MESSAGE)
        if not usager.identifiants and not usager.contacts:
            log.debug(self.IDENTIFIANTS_OU_CONTACTS_USAGER_PRESENT_MESSAGE)
            erreurs.append(self.IDENTIFIANTS_OU_CONTACTS_USAGER_PRESENT_MESSAGE)
        else:
            erreurs += self.__valider_champs_obligatoires_identifiants(usager.identifiants)
            erreurs += self.__valider_champs_obligatoires_contacts(usager.contacts)
        
        if erreurs:
            log.warn("L'usager n'est pas valide", extra={"erreurs":";".join(erreurs)})
        else:
            log.info("L'usager a été validé et n'a pas d'erreurs")

        return erreurs
    
    def __valider_champs_obligatoires_identifiants(self, identifiants: List[Usager.Identifiant]) -> List[str]:
        erreurs: List[str] = []
        
        if identifiants:
            for identifiant in identifiants:
                if not identifiant.type or not identifiant.valeur:
                    log.debug(self.TYPE_ET_VALEUR_IDENTIFIANTS_USAGER_PRESENTS_MESSAGE)
                    erreurs.append(self.TYPE_ET_VALEUR_IDENTIFIANTS_USAGER_PRESENTS_MESSAGE)
        
        return erreurs
    
    def __valider_champs_obligatoires_contacts(self, contacts: List[Usager.Contact]) -> List[str]:
        erreurs: List[str] = []
        
        if contacts:
            for contact in contacts:
                if not contact.type or not contact.prenom or not contact.nom:
                    log.debug(self.TYPE_ET_PRENOM_ET_NOM_CONTACTS_USAGER_PRESENTS_MESSAGE)
                    erreurs.append(self.TYPE_ET_PRENOM_ET_NOM_CONTACTS_USAGER_PRESENTS_MESSAGE)
        
        return erreurs

    def __valider_domaines_vocabulaire(self, usager: Usager) -> List[str]:
        erreurs: List[str] = []

        if usager.identifiants:
            for identifiant in usager.identifiants:
                if not (identifiant.type == self.TYPE_IDENTIFIANT_POUR_NIU or self._code_present_dans_domaine_vocabulaire(identifiant.type, self._value_set_client.get_identifiertype())):
                    log.debug(self.TYPE_IDENTIFIANT_INEXISTANT_MESSAGE.format(identifiant.type, identifiant.valeur))
                    erreurs.append(self.TYPE_IDENTIFIANT_INEXISTANT_MESSAGE.format(identifiant.type, identifiant.valeur))
                
        if usager.sexe and not self._code_present_dans_domaine_vocabulaire(usager.sexe, self._value_set_client.get_gender()):
            log.debug(self.SEXE_INEXISTANT_MESSAGE.format(usager.sexe))
            erreurs.append(self.SEXE_INEXISTANT_MESSAGE.format(usager.sexe))
        
        if usager.adresse:
            if usager.adresse.ville and not self._description_presente_dans_domaine_vocabulaire(usager.adresse.ville, self._value_set_client.get_city_by_name(name=usager.adresse.ville, province_code=usager.adresse.get_province_code())):
                log.debug(self.VILLE_INEXISTANTE_MESSAGE.format(usager.adresse.ville))
                erreurs.append(self.VILLE_INEXISTANTE_MESSAGE.format(usager.adresse.ville))
            if usager.adresse.province and not self._description_presente_dans_domaine_vocabulaire(usager.adresse.province, self._value_set_client.get_provinces()):
                log.debug(self.PROVINCE_INEXISTANTE_MESSAGE. format(usager.adresse.province))
                erreurs.append(self.PROVINCE_INEXISTANTE_MESSAGE. format(usager.adresse.province))
            if usager.adresse.pays and not self._description_presente_dans_domaine_vocabulaire(usager.adresse.pays, self._value_set_client.get_country()):
                log.debug(self.PAYS_INEXISTANT_MESSAGE.format(usager.adresse.pays))
                erreurs.append(self.PAYS_INEXISTANT_MESSAGE.format(usager.adresse.pays))
            
        if usager.contacts:
            for contact in usager.contacts:
                if not self._code_present_dans_domaine_vocabulaire(contact.type, self._value_set_client.get_contacttype()):
                    log.debug(self.TYPE_CONTACT_INEXISTANT_MESSAGE.format(contact.type, contact.prenom, contact.nom))
                    erreurs.append(self.TYPE_CONTACT_INEXISTANT_MESSAGE.format(contact.type, contact.prenom, contact.nom))
        
        return erreurs

class ValidateurUsagerPourAppariement:

    DATE_NAISSANCE_PRESENTE_MESSAGE = "L'usager doit avoir une date de naissance s'il n'a pas d'identifiants"
    SEXE_PRESENT_MESSAGE = "L'usager doit avoir un sexe s'il n'a pas d'identifiants"
    CONTACTS_PRESENT_MESSAGE = "L'usager doit avoir des contacts s'il n'a pas d'identifiants"

    def valider(self, usager: Usager) -> List[str]:
        erreurs: List[str] = []

        if not usager.identifiants:
            if not usager.prenom:
                log.debug(ValidateurUsager.PRENOM_USAGER_PRESENT_MESSAGE)
                erreurs.append(ValidateurUsager.PRENOM_USAGER_PRESENT_MESSAGE)
            if not usager.nom:
                log.debug(ValidateurUsager.NOM_USAGER_PRESENT_MESSAGE)
                erreurs.append(ValidateurUsager.NOM_USAGER_PRESENT_MESSAGE)
            if not usager.dateNaissance:
                log.debug(self.DATE_NAISSANCE_PRESENTE_MESSAGE)
                erreurs.append(self.DATE_NAISSANCE_PRESENTE_MESSAGE)
            if not usager.sexe:
                log.debug(self.SEXE_PRESENT_MESSAGE)
                erreurs.append(self.SEXE_PRESENT_MESSAGE)
            if not usager.contacts:
                log.debug(self.CONTACTS_PRESENT_MESSAGE)
                erreurs.append(self.CONTACTS_PRESENT_MESSAGE)
            else:
                erreurs += self.__valider_champs_obligatoires_contacts(usager.contacts)
        else:
            if not usager.prenom:
                log.debug(ValidateurUsager.PRENOM_USAGER_PRESENT_MESSAGE)
                erreurs.append(ValidateurUsager.PRENOM_USAGER_PRESENT_MESSAGE)
            if not usager.nom:
                log.debug(ValidateurUsager.NOM_USAGER_PRESENT_MESSAGE)
                erreurs.append(ValidateurUsager.NOM_USAGER_PRESENT_MESSAGE)
        
        return erreurs
    
    def __valider_champs_obligatoires_contacts(self, contacts) -> List[str]:
        erreurs: List[str] = []

        for contact in contacts:
            if not contact.type or not contact.prenom or not contact.nom:
                log.debug(ValidateurUsager.TYPE_ET_PRENOM_ET_NOM_CONTACTS_USAGER_PRESENTS_MESSAGE)
                erreurs.append(ValidateurUsager.TYPE_ET_PRENOM_ET_NOM_CONTACTS_USAGER_PRESENTS_MESSAGE)
        
        return erreurs


from fhirclient.models.condition import Condition
from fhirclient.models.extension import Extension
from fhirclient.models.coding import Coding
from inspqcommun.fhir.visitors.base import BaseVisitor

class ConditionVisitor(BaseVisitor):
    
    AGENT_URL = "http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/extensions/#condition/agent"
    ANTIGEN_URL = "http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/extensions/#condition/antigen"

    def __init__(self, fhir_resource=None) -> None:
        self.setFhirResource(fhir_resource if fhir_resource else Condition())
    
    def getFhirResource(self) -> Condition:
        return super().getFhirResource()

    def get_code(self) -> Coding:
        if self.getFhirResource().code:
            for coding in self.getFhirResource().code.coding:
                if coding.system == self.DEFAULT_CODING_SYSTEM and coding.version == self.DEFAULT_CODING_VERSION:
                    return coding
        return None

    def get_category(self) -> Coding:
        if self.getFhirResource().category:
            for coding in self.getFhirResource().category.coding:
                if coding.system == self.DEFAULT_CODING_SYSTEM and coding.version == self.DEFAULT_CODING_VERSION:
                    return coding
        return None
    
    def get_agent(self) -> Coding:
        return self.__get_extension_coding(url=self.AGENT_URL, system=self.DEFAULT_CODING_SYSTEM, version=self.DEFAULT_CODING_VERSION)

    def get_antigen(self) -> Coding:
        return self.__get_extension_coding(url=self.ANTIGEN_URL, system=self.DEFAULT_CODING_SYSTEM, version=self.DEFAULT_CODING_VERSION)

    def __get_extension_coding(self, url, system, version) -> Coding:
        if self.getFhirResource().extension is not None:
            extension: Extension
            coding: Coding
            for extension in self.getFhirResource().extension:
                if extension.url == url and extension.valueCodeableConcept is not None:
                    for coding in extension.valueCodeableConcept.coding:
                        if coding.system == system and coding.version == version:
                            return coding
        return None


from fhirclient.models.meta import Meta
from fhirclient.models.extension import Extension
from fhirclient.models.fhirdate import FHIRDate
from fhirclient.models.coding import Coding
from fhirclient.models.codeableconcept import CodeableConcept
from fhirclient.models.patient import Patient
from fhirclient.models.immunization import Immunization, ImmunizationExplanation
from fhirclient.models.location import Location
from fhirclient.models.practitioner import Practitioner
from fhirclient.models.quantity import Quantity
from fhirclient.models.reference import Reference

from inspqcommun.fhir.visitors.base import BaseVisitor

from datetime import date, datetime
from typing import List

class ImmunizationVisitor(BaseVisitor):
        
    OVERRIDE_STATUS_URL = "http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/extensions/#immunization/overridestatus"
    OVERRIDE_REASON_URL = "http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/extensions/#immunization/overridereason"
    ANTIGEN_STATUS_ANTIGEN_URL = "antigen"
    ANTIGEN_STATUS_STATUS_URL = "status"
    ANTIGEN_STATUS_DOSE_NUMER_URL = "doseNumber"
    LOT_ID_URL = "http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/extensions/#immunization/lotid"
    TRADE_NAME_URL = 'http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/extensions/#immunization/tradename'
    ANTIGEN_STATUS_URL = "http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/extensions/#immunization/antigenstatus"
    PROFILE_URL = 'http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/profiles/InspqImmunization.structuredefinition.xml'
    UPDATED_BY_URL = "http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/extensions/#common/updatedby"
    CREATED_BY_URL = 'http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/extensions/#common/createdby'
    CREATION_DATE_URL = 'http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/extensions/#common/creationdate'

    def __init__(self, fhir_resource=None) -> None:
        self.setFhirResource(fhir_resource=fhir_resource if fhir_resource else self.__build_default_immunization())
    
    def __build_default_immunization(self) -> Immunization:
        immunization = Immunization()
        immunization.wasNotGiven = False
        immunization.status = 'completed'
        return immunization

    def getFhirResource(self) -> Immunization:
        return super().getFhirResource()  
    
    def set_id(self, id=None):
        self.getFhirResource().id = id

    def get_id(self):
        return self.getFhirResource().id
        
    def get_meta_updated_by(self):
        if self.getFhirResource() is None or self.getFhirResource().meta is None or self.getFhirResource().meta.extension is None:
            return None
        for extension in self.getFhirResource().meta.extension:
            if extension.url == self.UPDATED_BY_URL:
                return extension.valueString
        return None

    def get_meta_created_by(self):
        if self.getFhirResource() is None or self.getFhirResource().meta is None or self.getFhirResource().meta.extension is None:
            return None
        for extension in self.getFhirResource().meta.extension:
            if extension.url == self.CREATED_BY_URL:
                return extension.valueString
        return None

    def get_meta_creation_date(self):
        if self.getFhirResource() is None or self.getFhirResource().meta is None or self.getFhirResource().meta.extension is None:
            return None
        for extension in self.getFhirResource().meta.extension:
            if extension.url == self.CREATION_DATE_URL:
                return extension.valueDate
        return None

    def get_meta_version_id(self):
        if self.getFhirResource() is None or self.getFhirResource().meta is None:
            return None
        return self.getFhirResource().meta.versionId

    def get_meta_last_updated(self):
        if self.getFhirResource() is None or self.getFhirResource().meta is None:
            return None
        if self.getFhirResource().meta.lastUpdated is not None:
            return self.getFhirResource().meta.lastUpdated.isostring
        else:
            return self.getFhirResource().meta.lastUpdated

    def get_meta_profile(self):
        if self.getFhirResource() is None or self.getFhirResource().meta is None:
            return None
        return self.getFhirResource().meta.profile
    
    def set_meta_updated_by(self, updated_by: str = None):
        if not self.getFhirResource().meta:
            self.getFhirResource().meta = Meta()
        if not self.getFhirResource().meta.extension:
            self.getFhirResource().meta.extension = []
        self.getFhirResource().meta.extension = super().add_or_update_extension_to_extensions(extension=self._creer_extension(url=self.UPDATED_BY_URL, valueString=updated_by), 
                                                                                              extensions=self.getFhirResource().meta.extension)
    def set_meta_last_updated(self, last_updated: date = None):
        if not self.getFhirResource().meta:
            self.getFhirResource().meta = Meta()
        self.getFhirResource().meta.lastUpdated = last_updated
    
    def set_meta_created_by(self, created_by: str = None):
        if not self.getFhirResource().meta:
            self.getFhirResource().meta = Meta()
        if not self.getFhirResource().meta.extension:
            self.getFhirResource().meta.extension = []
        self.getFhirResource().meta.extension = super().add_or_update_extension_to_extensions(extension=self._creer_extension(url=self.CREATED_BY_URL, valueString=created_by),
                                                                                              extensions=self.getFhirResource().meta.extension)
        
    def set_meta_creation_date(self, creation_date: date = None):
        if not self.getFhirResource().meta:
            self.getFhirResource().meta = Meta()
        if not self.getFhirResource().meta.extension:
            self.getFhirResource().meta.extension = []
        self.getFhirResource().meta.extension = super().add_or_update_extension_to_extensions(extension=self._creer_extension(url=self.CREATION_DATE_URL, valueDate=creation_date),
                                                                                              extensions=self.getFhirResource().meta.extension)
    
    def set_meta_profile(self, profile: str = None):
        if not self.getFhirResource().meta:
            self.getFhirResource().meta = Meta()
        self.getFhirResource().meta.profile = [ profile ]

    def set_meta_version_id(self, version_id: str = None):
        if not self.getFhirResource().meta:
            self.getFhirResource().meta = Meta()
        self.getFhirResource().meta.versionId = version_id

    def get_override_status(self) -> CodeableConcept:
        if self.getFhirResource().extension is not None:
            for ext in self.getFhirResource().extension:
                if ext.url == self.OVERRIDE_STATUS_URL:
                    return ext.valueCodeableConcept
        return None

    def set_override_status(self, status_code:str=None, status_display:str=None):
        status_codeable_concept = self._to_codeable_concept(
            code=status_code,
            display=status_display,
            coding_system=ImmunizationVisitor.DEFAULT_CODING_SYSTEM,
            coding_version=ImmunizationVisitor.DEFAULT_CODING_VERSION)
        self.getFhirResource().extension = super().add_or_update_extension_to_extensions(extension=self._creer_extension(self.OVERRIDE_STATUS_URL, valueCodeableConcept=status_codeable_concept), 
                                                                                         extensions=self.getFhirResource().extension)

    def get_override_reason(self) -> CodeableConcept:
        if self.getFhirResource().extension is not None:
            for ext in self.getFhirResource().extension:
                if ext.url == self.OVERRIDE_REASON_URL:
                    return ext.valueCodeableConcept
        return None

    def set_override_reason(self, status_code:str=None, status_display:str=None):
        status_codeable_concept = self._to_codeable_concept(
            code=status_code,
            display=status_display,
            coding_system=ImmunizationVisitor.DEFAULT_CODING_SYSTEM,
            coding_version=ImmunizationVisitor.DEFAULT_CODING_VERSION)
        self.getFhirResource().extension = super().add_or_update_extension_to_extensions(extension=self._creer_extension(self.OVERRIDE_REASON_URL, valueCodeableConcept=status_codeable_concept), 
                                                                                         extensions=self.getFhirResource().extension)
    def set_antigen_status(self, antigen_code:str=None, antigen_display:str=None, dose_number:int=None, status_code:str=None, status_display:str=None):
        antigen_status_ext = Extension()
        antigen_status_ext.url = self.ANTIGEN_STATUS_URL

        antigen_codeable_concept = self._to_codeable_concept(
            code=antigen_code,
            display=antigen_display,
            coding_system=ImmunizationVisitor.DEFAULT_CODING_SYSTEM,
            coding_version=ImmunizationVisitor.DEFAULT_CODING_VERSION)
        antigen_ext = self._creer_extension(url=self.ANTIGEN_STATUS_ANTIGEN_URL, valueCodeableConcept=antigen_codeable_concept)
        antigen_status_ext.extension = super().add_or_update_extension_to_extensions(extension=antigen_ext, extensions=antigen_status_ext.extension)

        dose_number_ext = self._creer_extension(url=self.ANTIGEN_STATUS_DOSE_NUMER_URL, valueInteger=dose_number)
        antigen_status_ext.extension = super().add_or_update_extension_to_extensions(extension=dose_number_ext, extensions=antigen_status_ext.extension)

        status_codeable_concept = self._to_codeable_concept(
            code = status_code,
            display=status_display,
            coding_system=ImmunizationVisitor.DEFAULT_CODING_SYSTEM,
            coding_version=ImmunizationVisitor.DEFAULT_CODING_VERSION
        )
        status_ext = self._creer_extension(url=self.ANTIGEN_STATUS_STATUS_URL, valueCodeableConcept=status_codeable_concept)
        antigen_status_ext.extension = super().add_or_update_extension_to_extensions(extension=status_ext, extensions=antigen_status_ext.extension)

        self.getFhirResource().extension = super().add_or_update_extension_to_extensions(extension=antigen_status_ext, extensions=self.getFhirResource().extension)

    def get_antigen_status(self) -> Extension:
        if self.getFhirResource().extension is not None:
            for ext in self.getFhirResource().extension:
                if ext.url == self.ANTIGEN_STATUS_URL:
                    return ext
        return None

    def get_antigen_status_antigen(self) -> Coding:
        if self.get_antigen_status() is not None:
            for antigen_extension in self.get_antigen_status().extension:
                if antigen_extension.url == self.ANTIGEN_STATUS_ANTIGEN_URL:
                    return self._get_coding_par_system(codeableConcept=antigen_extension.valueCodeableConcept,
                                                              coding_system=ImmunizationVisitor.DEFAULT_CODING_SYSTEM,
                                                              coding_version=ImmunizationVisitor.DEFAULT_CODING_VERSION)
        return None

    def get_antigen_status_dose_number(self) -> int:
        if self.get_antigen_status() is not None:
            for antigen_extension in self.get_antigen_status().extension:
                if antigen_extension.url == self.ANTIGEN_STATUS_DOSE_NUMER_URL:
                    return antigen_extension.valueInteger
        return None

    def get_antigen_status_status(self) -> Coding:
        if self.get_antigen_status() is not None:
            for antigen_extension in self.get_antigen_status().extension:
                if antigen_extension.url == self.ANTIGEN_STATUS_STATUS_URL:
                    return self._get_coding_par_system(codeableConcept=antigen_extension.valueCodeableConcept,
                                                              coding_system=ImmunizationVisitor.DEFAULT_CODING_SYSTEM,
                                                              coding_version=ImmunizationVisitor.DEFAULT_CODING_VERSION)
        return None

    def get_lot_id(self):
        if self.getFhirResource() is None or self.getFhirResource().extension is None:
            return None
        for extension in self.getFhirResource().extension:
            if extension.url == self.LOT_ID_URL:
                return extension.valueString
        return None
    
    def set_lot_id(self, lot_id=None):
        lot_id_ext = Extension()
        lot_id_ext.url = self.LOT_ID_URL
        lot_id_ext.valueString = lot_id
        self.getFhirResource().extension = super().add_or_update_extension_to_extensions(extension=lot_id_ext, extensions=self.getFhirResource().extension)

    def get_trade_name(self):
        if self.getFhirResource() is None or self.getFhirResource().extension is None:
            return None
        for extension in self.getFhirResource().extension:
            if extension.url == self.TRADE_NAME_URL:
                return extension.valueString
        return None

    def set_trade_name(self, trade_name=None):
        self.getFhirResource().extension = super().add_or_update_extension_to_extensions(extension=self._creer_extension(url=self.TRADE_NAME_URL, valueString=trade_name), 
                                                                                         extensions=self.getFhirResource().extension)
    
    def get_vaccine_code(self) -> Coding:
        vaccine_code = self._get_coding_par_system(
            codeableConcept=self.getFhirResource().vaccineCode)
        return vaccine_code

    def set_vaccine_code(self, vaccine_code:str=None, vaccine_display:str=None):
        self.getFhirResource().vaccineCode = self._to_codeable_concept(
            code=vaccine_code,
            display=vaccine_display)

    def get_patient(self) -> Reference:
        if self.getFhirResource().patient:
            return self.getFhirResource().patient        
        else:
            return None
    
    def set_patient(self, patient: Patient=None, patient_id: str=None):
        self.getFhirResource().patient = Reference()
        self.getFhirResource().patient.reference = patient.id if patient else patient_id

    def get_performer(self) -> Reference:
        if self.getFhirResource().performer is not None:
            return self.getFhirResource().performer
        return None

    def set_performer(self, performer: Practitioner = None, performer_id: str = None):
        self.getFhirResource().performer = Reference()
        self.getFhirResource().performer.reference = performer.id if performer else performer_id
        if performer and performer.name:
            self.getFhirResource().performer.display = performer.name.given[0] + ' ' + performer.name.family[0]

    def get_location(self) -> Reference:
        return self.getFhirResource().location if self.getFhirResource().location else None

    def set_location(self, location:Location=None, location_id: str = None):
        self.getFhirResource().location = Reference()
        self.getFhirResource().location.reference = location.id if location else location_id
        if location:
            self.getFhirResource().location.display = location.name

    def get_lot_number(self) -> str:
        return self.getFhirResource().lotNumber

    def set_lot_number(self, lot_number=None):
        self.getFhirResource().lotNumber = lot_number

    def get_expiration_date(self) -> FHIRDate:
        if self.getFhirResource().expirationDate is not None:
            return self.getFhirResource().expirationDate
        else:
            return None

    def set_expiration_date(self, expiration_date=None):
        if type(expiration_date) is str:
            self.getFhirResource().expirationDate = datetime.strptime(expiration_date, "%Y-%m-%d")
        else:
            self.getFhirResource().expirationDate = expiration_date

    def get_site(self) -> Coding:
        if self.getFhirResource().site is not None:
            coding = self._get_coding_par_system(
                codeableConcept=self.getFhirResource().site)
            return coding
        return None

    def set_site(self, site_code=None, site_display=None):
        self.getFhirResource().site = self._to_codeable_concept(
            code=site_code,
            display=site_display
        )

    def get_route(self) -> Coding:
        if self.getFhirResource().route is not None:
            coding = self._get_coding_par_system(
                codeableConcept=self.getFhirResource().route)
            return coding
        return None

    def set_route(self, route_code=None, route_display=None):
        self.getFhirResource().route = self._to_codeable_concept(
            code=route_code,
            display=route_display
        )

    def get_dose_quantity(self) -> Quantity:
        return self.getFhirResource().doseQuantity
    
    def set_dose_quantity(self, quantity=None, quantity_code=None, quantity_unit=None, quantity_value=None, quantity_system=None):
        if quantity is not None:
            dose_quantity = quantity
        else:
            dose_quantity = Quantity()
            dose_quantity.code = quantity_code
            dose_quantity.system = quantity_system if quantity_system else ImmunizationVisitor.DEFAULT_CODING_SYSTEM
            dose_quantity.unit = quantity_unit
            dose_quantity.value = quantity_value

        self.getFhirResource().doseQuantity = dose_quantity

    def get_reasons(self) -> List[Coding]:
        if self.getFhirResource().explanation is not None and self.getFhirResource().explanation.reason is not None:
            reasons = []
            for reason in self.getFhirResource().explanation.reason:
                if type(reason) is CodeableConcept:
                    reason_coding = self._get_coding_par_system(
                        codeableConcept=reason)
                    reasons.append(reason_coding)
                elif type(reason) is Coding:
                    reasons.append(reason)
            return reasons

    def set_reason(self, reason_code:str=None, reason_display:str=None):
        self.getFhirResource().explanation = ImmunizationExplanation()
        self.getFhirResource().explanation.reason = []
        self.getFhirResource().explanation.reason.append(self._to_codeable_concept(code=reason_code,
                                                                                    display=reason_display))

    def get_date(self) -> FHIRDate:
        return self.getFhirResource().date

    def set_date(self, date_to_set=None):
        if date_to_set is not None:
            if type(date_to_set) is date or type(date_to_set) is datetime:
                date_to_set = datetime.strftime(date_to_set, "%Y-%m-%dT%H:%M:%S")
        self.getFhirResource().date = FHIRDate(jsonval=date_to_set) if date_to_set is not None else None   

    def get_status(self) -> str:
        return self.getFhirResource().status

    def set_status(self, immunization=None, status=None):
        if immunization is None:
            immunization = self.fhir_resource
        if status is not None:
            self.status = status
        immunization.status = self.status


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import io
import os

from setuptools import setup, find_packages

def get_version():
    output_stream = os.popen('git describe')
    version = output_stream.read().strip()
    splitted_version = version.split('-')
    final_version = splitted_version[0]
    if len(splitted_version) > 1:
        final_version = "{version}.post{additionnal_commits}".format(
            version=splitted_version[0],
            additionnal_commits=splitted_version[1])
    return final_version

setup(
    # Needed to silence warnings (and to be a worthwhile package)
    name='inspqcommun',
    url='https://gitlab.forge.gouv.qc.ca/inspq/commun/python/inspqcommun.git',
    author='Philippe Gauthier',
    author_email='philippe.gauthier@inspq.qc.ca',
    # Needed to actually package something
    packages=find_packages(include=['inspqcommun*']),
    # Needed for dependencies
    install_requires=['fhirclient==1.0.3','wheel','urllib3','requests','pyjwt','jinja2','PyYAML','confluent_kafka==1.8.2','pygelf','six','str2bool'],
    # *strongly* suggested for sharing
    version=get_version(),
    #version_command=('git describe', "pep440-git-local"),
    # The license can be anything you like
    license='LiLiQ',
    description='Librairies communes de INSPQ',
    long_description=io.open('README.md', 'r', encoding="utf-8").read(),
)



from fhirclient.models.valueset import ValueSet
from inspqcommun.fhir.visitors.base import BaseVisitor

class ValueSetVisitor(BaseVisitor):
    
    def __init__(self, fhir_resource=None) -> None:
        super().setFhirResource(fhir_resource if fhir_resource else ValueSet())

    def getFhirResource(self) -> ValueSet:
        return super().getFhirResource()

    def search_code_system_concept(self, code=None, display=None):
        code_system_concept = None
        if (self.getFhirResource() is not None
                and self.getFhirResource().codeSystem is not None
                and self.getFhirResource().codeSystem.concept is not None
                and len(self.getFhirResource().codeSystem.concept) > 0):
            for concept in self.getFhirResource().codeSystem.concept:
                if ((code is not None and concept.code == code)
                        or (display is not None and concept.display == display)):
                    code_system_concept = concept

        return code_system_concept
    
    def find_code_system_concept_by_code(self, code: str) -> bool:
        return True if self.search_code_system_concept(code=code) else False
    
    def find_code_system_concept_by_display(self, display: str) -> bool:
        return True if self.search_code_system_concept(display=display) else False


from fhirclient.models.codeableconcept import CodeableConcept
from fhirclient.models.coding import Coding

class ConvertisseurBase:
    
    def _convertirEnCodeableConcept(self, code: str) -> CodeableConcept:
        codeableConcept = CodeableConcept()
        coding = Coding()
        coding.code = code
        coding.system = 'http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/vocabulary'
        coding.version = '1.0.0'
        codeableConcept.coding = [ coding ]
        return codeableConcept


import json

class Ressource:
    
    ressource: str

    def __init__(self, ressource: str) -> None:
        self.ressource = ressource

    def as_json(self) -> str:
        return json.dumps(self, default=lambda o: o.__dict__ if hasattr(o, '__dict__') else str(o), sort_keys=True, indent=4)

    def _get_value_from_kwargs(self, key: str, clazz=None, defaultValue=None, **kwargs):
        if clazz:
            return clazz(**kwargs[key]) if kwargs and key in kwargs else defaultValue
        else:
            return kwargs[key] if kwargs and key in kwargs else defaultValue
    
    def _get_list_of_values_from_kwargs(self, key: str, clazz=None, defaultValue=None, **kwargs):
        if kwargs and key in kwargs:
            if clazz:
                return [ clazz(**valeur) for valeur in kwargs[key] ]
            else:
                return [ valeur for valeur in kwargs[key] ]
        else:
            return defaultValue if defaultValue else []


from inspqcommun.fa.ressources.usager import Usager
from inspqcommun.fa.ressources.acte_vaccinal import ActeVaccinal
from inspqcommun.fhir.clients.patient_client import PatientClient
from inspqcommun.fa.convertisseurs.convertisseur_usager import ConvertisseurUsager
from inspqcommun.fa.convertisseurs.convertisseur_base import ConvertisseurBase

from fhirclient.models.extension import Extension
from fhirclient.models.immunization import Immunization, ImmunizationExplanation
from fhirclient.models.quantity import Quantity
from fhirclient.models.reference import Reference
from fhirclient.models.annotation import Annotation

import logging

log = logging.getLogger(__name__)

class ConvertisseurActeVaccinal(ConvertisseurBase):

    def __init__(self, convertisseur_usager: ConvertisseurUsager, patient_client: PatientClient) -> None:
        self.__convertisseur_usager : ConvertisseurUsager = convertisseur_usager
        self.__patient_client: PatientClient = patient_client

    def fromJson(self, **kwargs):
        return ActeVaccinal(**kwargs)

    def toFhir(self, acte_vaccinal: ActeVaccinal) -> Immunization:
        log.debug("Acte vaccinal à convertir: {}".format(acte_vaccinal.as_json()))

        immunization = Immunization()

        immunization.status = 'completed'
        immunization.wasNotGiven = False

        immunization.patient = self.__obtenir_reference_usager_par_appariement(acte_vaccinal.usager)

        immunization.extension = []

        if acte_vaccinal.nomCommercial:
            tradename = Extension()
            tradename.url = 'http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/extensions/#immunization/tradename'
            tradename.valueString = acte_vaccinal.nomCommercial
            immunization.extension.append(tradename)

            if acte_vaccinal.lot:
                lotId = Extension()
                lotId.url = 'http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/extensions/#immunization/lotid'
                lotId.valueString = acte_vaccinal.lot.id
                immunization.extension.append(lotId)
                immunization.lotNumber = acte_vaccinal.lot.numero
        
        if acte_vaccinal.agent:
            immunization.vaccineCode = self._convertirEnCodeableConcept(acte_vaccinal.agent)

        immunization.date = acte_vaccinal.dateAdministration
        
        if acte_vaccinal.quantiteAdministree:
            immunization.doseQuantity = Quantity()
            immunization.doseQuantity.value = acte_vaccinal.quantiteAdministree.quantite
            immunization.doseQuantity.code = acte_vaccinal.quantiteAdministree.unite if acte_vaccinal.quantiteAdministree.unite else 'I'
            immunization.doseQuantity.system = 'http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/vocabulary'

        immunization.route = self._convertirEnCodeableConcept(acte_vaccinal.voieAdministration) if acte_vaccinal.voieAdministration else self._convertirEnCodeableConcept('UNKN')
        immunization.site = self._convertirEnCodeableConcept(acte_vaccinal.siteAdministration) if acte_vaccinal.siteAdministration else self._convertirEnCodeableConcept('I')

        if acte_vaccinal.raisonAdministration:
            immunization.explanation = []
            explanation = ImmunizationExplanation()
            explanation.reason = [ self._convertirEnCodeableConcept(acte_vaccinal.raisonAdministration) ]
            immunization.explanation.append(explanation)

        immunization.performer = Reference()
        immunization.performer.reference = str(acte_vaccinal.vaccinateur.id) if acte_vaccinal.vaccinateur else '20497'

        immunization.location = Reference()
        immunization.location.reference = str(acte_vaccinal.lieuVaccination.id) if acte_vaccinal.lieuVaccination else '6906'

        if acte_vaccinal.commentaires:
            immunization.note = []
            for commentaire in acte_vaccinal.commentaires:
                note = Annotation()
                note.authorString = commentaire.auteur
                note.time = commentaire.date
                note.text = commentaire.texte
                immunization.note.append(note)

        immunization.reported = not acte_vaccinal.nomCommercial or not acte_vaccinal.quantiteAdministree

        log.debug("Immunization obtenue suite à la conversion: {}".format(immunization.as_json()))
        return immunization
    
    def __obtenir_reference_usager_par_appariement(self, usager: Usager) -> Reference:
        patient = self.__convertisseur_usager.toFhir(usager)
        response = self.__patient_client.match(patient)
        if response.status_code == 200:
            patient = self.__patient_client.extract_patient_from_response(response)
            log.debug("Patient reçu pour l'ajout d'un acte vaccinal suite à un appariement: {}".format(patient.as_json()))
            reference = Reference()
            reference.reference = str(patient.id)
            return reference
        else:
            return None


from inspqcommun.fa.ressources.usager import Usager
from inspqcommun.fa.convertisseurs.convertisseur_base import ConvertisseurBase
from fhirclient.models.patient import Patient, PatientContact
from fhirclient.models.identifier import Identifier
from fhirclient.models.humanname import HumanName
from fhirclient.models.extension import Extension
from fhirclient.models.address import Address
from fhirclient.models.contactpoint import ContactPoint
from fhirclient.models.fhirdate import FHIRDate
import logging

log = logging.getLogger(__name__)

class ConvertisseurUsager(ConvertisseurBase):

    def __init__(self) -> None:
        self.__convertisseurAdresse = ConvertisseurUsager.ConvertisseurAdresse()
        self.__convertisseurTelephone = ConvertisseurUsager.ConvertisseurTelephone()
        self.__convertisseurContact = ConvertisseurUsager.ConvertisseurContact()

    def fromJson(self, **kwargs):
        return Usager(**kwargs)

    def toFhir(self, usager: Usager) -> Patient:
        log.debug("Usager à convertir: {}".format(usager.as_json()))

        patient = Patient()
        patient.active = True
        
        patient.identifier = []
        if usager.identifiants:
            for identifiant in usager.identifiants:
                patient.identifier.append(self.__convertirIdentifiant(identifiant))
        else:
            patient.identifier.append(self.__convertirIdentifiant(Usager.Identifiant(type='AUCUN', valeur=None)))
        
        if usager.prenom or usager.nom:
            patient.name = [ self.__convertirPrenomNom(usager.prenom, usager.nom) ]
        
        patient.birthDate = FHIRDate(jsonval=usager.dateNaissance) if usager.dateNaissance is not None else None
        patient.gender = usager.sexe

        if usager.adresse:
            patient.address = [ self.__convertisseurAdresse.toFhir(usager.adresse) ]

        if usager.telephone and usager.telephone.numero:
            patient.telecom = [ self.__convertisseurTelephone.toFhir(usager.telephone) ]
        
        if usager.contacts:
            patient.contact = []
            for contact in usager.contacts:
                patient.contact.append(self.__convertisseurContact.toFhir(contact))

        log.debug("Patient obtenu suite à la conversion: {}".format(patient.as_json()))
        return patient

    def __convertirIdentifiant(self, identifiant: Usager.Identifiant) -> Identifier:
        identifier = Identifier()
        identifier.value = identifiant.valeur
        if identifiant.type:
            identifier.system = "http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/vocabulary/identifierType?code={}".format(identifiant.type)
        identifier.type = self._convertirEnCodeableConcept(identifiant.type)
        if identifiant.type == 'NAM':
            extension = Extension()
            extension.url = 'http://www.santepublique.rtss.qc.ca/sipmi/fa/1.0.0/extensions/#patient/healthcardorigin'
            extension.valueCodeableConcept = self._convertirEnCodeableConcept('QC')
            identifier.extension = [ extension ]
        return identifier

    def __convertirPrenomNom(self, prenom: str, nom: str) -> HumanName:
        name = HumanName()
        name.given = [ prenom ]
        name.family = [ nom ]
        return name
    
    class ConvertisseurAdresse(ConvertisseurBase):

        def toFhir(self, adresse: Usager.Adresse) -> Address:
            address = Address()
            address.line = [ adresse.adresse ]
            address.city = adresse.ville
            address.state = adresse.province
            address.country = adresse.pays
            address.postalCode = adresse.codePostal
            return address
        
    class ConvertisseurTelephone(ConvertisseurBase):

        def toFhir(self, telephone: Usager.Telephone) -> ContactPoint:
            telecom = ContactPoint()
            telecom.system = 'phone'
            telecom.value = '+1' + telephone.numero
            if telephone.extension:
                telecom.value = telecom.value + '#' + telephone.extension
            return telecom
        
    class ConvertisseurContact(ConvertisseurBase):

        def toFhir(self, contact: Usager.Contact) -> PatientContact:
            patientContact = PatientContact()
            patientContact.relationship = [ self._convertirEnCodeableConcept(contact.type) ]
            patientContact.name = HumanName()
            patientContact.name.given = [ contact.prenom ]
            patientContact.name.family = [ contact.nom ]
            return patientContact